﻿--[[---------------------------------------------------------------------------

Copyright (c) 2008 by K. Scott Piel 
All Rights Reserved

E-mail: < kscottpiel@gmail.com >
Web:    < http://www.scottpiel.com >

This file is part of nUI.

    nUI is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    nUI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with nUI.  If not, see <http://www.gnu.org/licenses/>.
	
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

--]]---------------------------------------------------------------------------

-- With thanks to Mairelon for whose secure state driver tutorial was the 
-- inspiration for this code and Alestane who helped me figure out how to
-- support the possession bar

-- basic layout

nUI_ActionBarLayout =
{
	btn_size = 55,
	gap      = 1,
	anchor   = "BOTTOM",
	xOfs     = 0,
	yOfs     = 122,
};

-- the Action bar frame

nUI_ActionBar = CreateFrame( "Frame", "nUI_ActionBar", nUI_Dashboard, "SecureStateHeaderTemplate" );
nUI_ActionBar_Events = CreateFrame( "Frame", "$parent_Events", nUI_ActionBar );

nUI_ActionBar:SetBackdrop(
	{
		bgFile   = "Interface\\AddOns\\nUI\\Art\\nUI_BevelboxBg.blp", 
		edgeFile = "Interface\\AddOns\\nUI\\Art\\nUI_BevelboxBorder.blp", 
		tile     = true, 
		tileSize = 1, 
		edgeSize = 12, 
		insets   = {left = 0, right = 0, top = 0, bottom = 0},
	}
);

nUI_ActionBar:SetBackdropColor( 0, 0, 0, 0.35 );

-- button initialization

nUI_ActionBar.Buttons = {};

for i=1,12 do

	local name       = "nUI_ActionButton"..i;
	local button     = CreateFrame( "CheckButton", "$parent_Button"..i, nUI_ActionBar, "ActionBarButtonTemplate" );
	local key1, key2 = GetBindingKey( "ACTIONBUTTON"..i );
	
	nUI_ActionBar.Buttons[i] = button;

	button:SetID( i );
	button:SetParent( nUI_ActionBar );
	button:SetNormalTexture( "" );
	button.SetNormalTexture = function() end;

	button.actionType               = "ACTIONBUTTON";
	button.actionID                 = i;
	button.nUI_CachedClearAllPoints = button.ClearAllPoints;
	button.nUI_CachedSetAllPoints   = button.SetAllPoints;
	button.nUI_CachedSetPoint       = button.SetPoint;

	button.ClearAllPoints = function() end;
	button.SetAllPoints   = function() end;
	button.SetPoint       = function() end;
	
	nUI_ActionBar:SetAttribute( "addchild", button );
		
end

-- set up the action bar state driver

nUI_ActionBar:SetAttribute( "stateattributes", "2:pg2;3:pg3;4:pg4;5:pg5;6:pg6;7:pg7;8:pg8;9:pg9;10:pg10;11:pg11;*:pg1" );
nUI_ActionBar:SetAttribute( "attributes-pg1",  "actionpage=1" );
nUI_ActionBar:SetAttribute( "attributes-pg2",  "actionpage=2" );
nUI_ActionBar:SetAttribute( "attributes-pg3",  "actionpage=3" );
nUI_ActionBar:SetAttribute( "attributes-pg4",  "actionpage=4" );
nUI_ActionBar:SetAttribute( "attributes-pg5",  "actionpage=5" );
nUI_ActionBar:SetAttribute( "attributes-pg6",  "actionpage=6" );
nUI_ActionBar:SetAttribute( "attributes-pg7",  "actionpage=7" );
nUI_ActionBar:SetAttribute( "attributes-pg8",  "actionpage=8" );
nUI_ActionBar:SetAttribute( "attributes-pg9",  "actionpage=9" );
nUI_ActionBar:SetAttribute( "attributes-pg10", "actionpage=10" );
nUI_ActionBar:SetAttribute( "attributes-pg11", "actionpage=11" );
nUI_ActionBar:SetAttribute( "statemap-page",   "*:$input" );

RegisterStateDriver( nUI_ActionBar, "page", "[bar:2] 2; [bar:3] 3; [bar:4] 4; [bar:5] 5; [bar:6] 6; [bonusbar:1] 7; [bonusbar:2] 8; [bonusbar:3] 9; [bonusbar:5] 11; 1" );

-- manage the layout of the Action Bar

nUI_ActionBar_Events:SetScript( "OnEvent", function() nUI_ActionBar_Events:layoutFrame(); end );
nUI_ActionBar_Events:RegisterEvent( "VARIABLES_LOADED" );
nUI_ActionBar_Events:RegisterEvent( "DISPLAY_SIZE_CHANGED" );

nUI_ActionBar_Events.layoutFrame = function()

	nUI:setScale();
				
	if nUI_ActionBar.fCache then nUI:lockFrame( nUI_ActionBar, false ); end
	
	local btn_size    = nUI_ActionBarLayout.btn_size * nUI.scale;
	local btn_gap     = nUI_ActionBarLayout.gap * nUI.scale;
	local width       = 0;
	local last_button;
	
	nUI_ActionBar:SetScale( 1.0 / nUI.scale );	
	nUI_ActionBar:SetFrameStrata( nUI_Dashboard:GetFrameStrata() );
	nUI_ActionBar:SetFrameLevel( nUI_Dashboard:GetFrameLevel()+2 );

	ActionBarUpButton:SetFrameStrata( nUI_ActionBar:GetFrameStrata() );
	ActionBarUpButton:SetFrameLevel( nUI_ActionBar:GetFrameLevel() );
	ActionBarUpButton:SetScale( 1.5 * nUI.scale );
	ActionBarUpButton:ClearAllPoints();
	ActionBarUpButton:SetPoint( "BOTTOMRIGHT", nUI_ActionBar, "BOTTOMLEFT", 10 * nUI.scale, -15 * nUI.scale );

	ActionBarDownButton:SetFrameStrata( nUI_ActionBar:GetFrameStrata() );
	ActionBarDownButton:SetFrameLevel( nUI_ActionBar:GetFrameLevel() );
	ActionBarDownButton:SetScale( 1.5 * nUI.scale );
	ActionBarDownButton:ClearAllPoints();
	ActionBarDownButton:SetPoint( "BOTTOMLEFT", nUI_ActionBar, "BOTTOMRIGHT", -10 * nUI.scale, -15 * nUI.scale  );
	
	for i in pairs( nUI_ActionBar.Buttons ) do
		
		local button = nUI_ActionBar.Buttons[i];

		button:SetFrameStrata( nUI_ActionBar:GetFrameStrata() );
		button:SetFrameLevel( nUI_ActionBar:GetFrameLevel()+1 );
		button:SetScale( btn_size / button:GetHeight() );

		width = width + button:GetWidth() * button:GetEffectiveScale() + (i > 1 and btn_gap or 0);
		
		button:nUI_CachedClearAllPoints();
		
		if i == 1 then 
			button:nUI_CachedSetPoint( "TOPLEFT", nUI_ActionBar, "TOPLEFT", btn_size * 0.05, -btn_size * 0.05 );
		else 
			button:nUI_CachedSetPoint( "TOPLEFT", last_button, "TOPRIGHT", btn_gap, 0 );
		end
				
		nUI:initActionButton( button );	

		last_button = button;
	end		
	
	nUI_ActionBar:SetHeight( btn_size * 1.1 );
	nUI_ActionBar:SetWidth( width + btn_size * 0.1 );
	nUI_ActionBar:SetPoint( nUI_ActionBarLayout.anchor, nUI_Dashboard, "CENTER", nUI_ActionBarLayout.xOfs * nUI.scale, nUI_ActionBarLayout.yOfs * nUI.scale );
	
	nUI:lockFrame( nUI_ActionBar, true, "Action Bar / Page 1" );
end
