﻿--[[---------------------------------------------------------------------------

Copyright (c) 2008 by K. Scott Piel 
All Rights Reserved

E-mail: < kscottpiel@gmail.com >
Web:    < http://www.scottpiel.com >

This file is part of nUI.

    nUI is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    nUI is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with nUI.  If not, see <http://www.gnu.org/licenses/>.
	
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR 
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE 
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

--]]---------------------------------------------------------------------------
	
-------------------------------------------------------------------------------
-- cache

local ActionHasRange        = ActionHasRange;
local CreateFrame           = CreateFrame;
local InCombatLockdown      = InCombatLockdown;
local CheckInteractDistance = CheckInteractDistance;
local UnitIsUnit            = UnitIsUnit;
local UnitExists            = UnitExists;
local UnitIsFriend          = UnitIsFriend;
local UnitOnTaxi            = UnitOnTaxi;
local UnitIsDeadOrGhost     = UnitIsDeadOrGhost;
local UnitCanAttack         = UnitCanAttack;
local UnitPlayerControlled  = UnitPlayerControlled;
local UnitPowerType         = UnitPowerType;
local UnitManaMax           = UnitManaMax;
local UnitMana              = UnitMana;
local UnitClassification    = UnitClassification;
local UnitIsVisible         = UnitIsVisible;
local GetSpellName          = GetSpellName;
local GetSpellInfo          = GetSpellInfo;
local SpellHasRange         = SpellHasRange;
local IsSpellInRange        = IsSpellInRange;
local SpellStatus           = nUI.SpellStatus;
local Cooldowns             = nUI.Cooldowns;

-------------------------------------------------------------------------------
-- spells we manualy exclude from the range check logic

local nUI_ExcludedRangedSpells =
{
	["Raptor Strike"] = 1,
	["Dismiss Pet"] = 1,
	["Eagle Eye"] = 1,
	["Fishing"] = 1,
	["Eyes of the Beast"] = 1,
	["Hunter's Mark"] = 1,
	["Misdirection"] = 1,
	["Flare"] = 1,
};

-------------------------------------------------------------------------------
-- spells we know to be zero max range but valid for range checking

local nUI_ZeroRangeSpells =
{
	["Mongoose Bite"] = 1,
	["Wing Clip"] = 1,
};

-------------------------------------------------------------------------------
-- used to sort the list of cooldowns in the hud from first expiration to last

local function nUI_SortCooldowns( left, right )

	local result = false;
	
	if left.ends < right.ends then
		result = true;
	end
	
	return result;
end

-------------------------------------------------------------------------------

nUI.Class.HUD = nUI.Class:CreatePrototype();
local prototype  = nUI.Class.HUD.prototype;

-------------------------------------------------------------------------------

function prototype:init( parent, id, layout )
	
	self.parent          = parent;
	self.id              = id;
	self.rid             = 1;
	self.Timers          = {};
	self.Timers.update   = 0;
	self.Timers.range    = 0;
	self.Timers.cooldown = 0;
	self.RangeList       = {};

	self:layoutFrame( layout, true );
	
end

-------------------------------------------------------------------------------

function prototype:layoutFrame( layout, is_init )

	if self.frame then 
		self:unregisterEvents(); 
		self.frame:SetAlpha( 1 );
		self.frame:Hide();
	end
	
	self.layout          = layout;
	self.scale   = nUI.scale * self.layout.scale;
	local height = 512 * self.scale;
	local width  = 128 * self.scale;

	-- if we haven't created this object yet, create its fram elements
	
	if is_init then
		
		-- player data elements
		
		self.frame                    = CreateFrame( "Frame", "nUI_HUD", self.parent );
		self.frame.player             = CreateFrame( "Frame", "$parent_Player", self.frame );
		self.frame.player.mana        = CreateFrame( "Frame", "$parentMana", self.frame.player );
		self.frame.player.rage        = CreateFrame( "Frame", "$parentRage", self.frame.player );
		self.frame.player.energy      = CreateFrame( "Frame", "$parentEnergy", self.frame.player );
		self.frame.player.focus       = CreateFrame( "Frame", "$parentFocus", self.frame.player );
		self.frame.player.health      = CreateFrame( "Frame", "$parentHealth", self.frame.player );
		self.frame.player.raid_icon   = CreateFrame( "Frame", "$parentRaidIcon", self.frame.player );
		self.frame.player.powerframe  = CreateFrame( "Frame", "$parentPowerFrame", self.frame.player );
		self.frame.player.healthframe = CreateFrame( "Frame", "$parentHealthFrame", self.frame.player );

		self.frame.player.mana.texture        = self.frame.player.mana:CreateTexture( "$parentTexture" );
		self.frame.player.rage.texture        = self.frame.player.rage:CreateTexture( "$parentTexture" );
		self.frame.player.energy.texture      = self.frame.player.energy:CreateTexture( "$parentTexture" );
		self.frame.player.focus.texture       = self.frame.player.focus:CreateTexture( "$parentTexture" );
		self.frame.player.health.texture      = self.frame.player.health:CreateTexture( "$parentTexture" );
		self.frame.player.raid_icon.texture   = self.frame.player.raid_icon:CreateTexture( "$parentTexture" );
		self.frame.player.powerframe.texture  = self.frame.player.powerframe:CreateTexture( "$parentTexture" );
		self.frame.player.healthframe.texture = self.frame.player.healthframe:CreateTexture( "$parentTexture" );

		self.frame.player.mana.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\lpb.blp" );
		self.frame.player.mana.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );

		self.frame.player.rage.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\lpb.blp" );
		self.frame.player.rage.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.player.energy.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\lpb.blp" );
		self.frame.player.energy.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );

		self.frame.player.focus.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\lpb.blp" );
		self.frame.player.focus.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );

		self.frame.player.health.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\lhb.blp" );
		self.frame.player.health.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );

		self.frame.player.raid_icon.texture:SetTexture( "Interface\\TargetingFrame\\UI-RaidTargetingIcons" );
		
		self.frame.player.powerframe.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\lpf.blp" );
		self.frame.player.powerframe.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
	
		self.frame.player.healthframe.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\lhf.blp" );
		self.frame.player.healthframe.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.player.unit = "player";
		self.frame.player:SetAttribute( "unit", "player" );
		RegisterUnitWatch( self.frame.player );
		
		-- pet data elements
		
		self.frame.pet             = CreateFrame( "Frame", "$parent_Pet", self.frame );
		self.frame.pet.health      = CreateFrame( "Frame", "$parentHealth", self.frame.pet );
		self.frame.pet.mana        = CreateFrame( "Frame", "$parentMana", self.frame.pet );
		self.frame.pet.rage        = CreateFrame( "Frame", "$parentRage", self.frame.pet );
		self.frame.pet.energy      = CreateFrame( "Frame", "$parentEnergy", self.frame.pet );
		self.frame.pet.focus       = CreateFrame( "Frame", "$parentFocus", self.frame.pet );
		self.frame.pet.raid_icon   = CreateFrame( "Frame", "$parentRaidIcon", self.frame.pet );
		self.frame.pet.healthframe = CreateFrame( "Frame", "$parentHealthFrame", self.frame.pet );
		self.frame.pet.powerframe  = CreateFrame( "Frame", "$parentPowerFrame", self.frame.pet );

		self.frame.pet.health.texture      = self.frame.pet.health:CreateTexture( "$parentTexture" );
		self.frame.pet.mana.texture        = self.frame.pet.mana:CreateTexture( "$parentTexture" );
		self.frame.pet.rage.texture        = self.frame.pet.rage:CreateTexture( "$parentTexture" );
		self.frame.pet.energy.texture      = self.frame.pet.energy:CreateTexture( "$parentTexture" );
		self.frame.pet.focus.texture       = self.frame.pet.focus:CreateTexture( "$parentTexture" );
		self.frame.pet.raid_icon.texture   = self.frame.pet.raid_icon:CreateTexture( "$parentTexture" );
		self.frame.pet.healthframe.texture = self.frame.pet.healthframe:CreateTexture( "$parentTexture" );
		self.frame.pet.powerframe.texture  = self.frame.pet.powerframe:CreateTexture( "$parentTexture" );
		
		self.frame.pet.health.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\lh2b.blp" );
		self.frame.pet.health.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );

		self.frame.pet.mana.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\lp2b.blp" );
		self.frame.pet.mana.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.pet.rage.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\lp2b.blp" );
		self.frame.pet.rage.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.pet.energy.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\lp2b.blp" );
		self.frame.pet.energy.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.pet.focus.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\lp2b.blp" );
		self.frame.pet.focus.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.pet.raid_icon.texture:SetTexture( "Interface\\TargetingFrame\\UI-RaidTargetingIcons" );
		
		self.frame.pet.healthframe.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\lh2f.blp" );
		self.frame.pet.healthframe.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.pet.powerframe.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\lp2f.blp" );
		self.frame.pet.powerframe.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.pet.unit = "pet";
		self.frame.pet:SetAttribute( "unit", "pet" );
		RegisterUnitWatch( self.frame.pet );
		
		-- target data elements
				
		self.frame.target             = CreateFrame( "Frame", "$parent_Target", self.frame );
		self.frame.target.mana        = CreateFrame( "Frame", "$parentMana", self.frame.target );
		self.frame.target.rage        = CreateFrame( "Frame", "$parentRage", self.frame.target );
		self.frame.target.energy      = CreateFrame( "Frame", "$parentEnergy", self.frame.target );
		self.frame.target.focus       = CreateFrame( "Frame", "$parentFocus", self.frame.target );
		self.frame.target.health      = CreateFrame( "Frame", "$parentHealth", self.frame.target );
		self.frame.target.raid_icon   = CreateFrame( "Frame", "$parentRaidIcon", self.frame.target );
		self.frame.target.powerframe  = CreateFrame( "Frame", "$parentPowerFrame", self.frame.target );
		self.frame.target.healthframe = CreateFrame( "Frame", "$parentHealthFrame", self.frame.target );

		self.frame.target.mana.texture        = self.frame.target.mana:CreateTexture( "$parentTexture" );
		self.frame.target.rage.texture        = self.frame.target.rage:CreateTexture( "$parentTexture" );
		self.frame.target.energy.texture      = self.frame.target.energy:CreateTexture( "$parentTexture" );
		self.frame.target.focus.texture       = self.frame.target.focus:CreateTexture( "$parentTexture" );
		self.frame.target.health.texture      = self.frame.target.health:CreateTexture( "$parentTexture" );
		self.frame.target.raid_icon.texture   = self.frame.target.raid_icon:CreateTexture( "$parentTexture" );
		self.frame.target.powerframe.texture  = self.frame.target.powerframe:CreateTexture( "$parentTexture" );
		self.frame.target.healthframe.texture = self.frame.target.healthframe:CreateTexture( "$parentTexture" );
	
		self.frame.target.combo_pt = {};
		
		self.frame.target.combo_pt[1] = CreateFrame( "Frame", "$parentComboPt"..1, self.frame.target );		
		self.frame.target.combo_pt[2] = CreateFrame( "Frame", "$parentComboPt"..2, self.frame.target );		
		self.frame.target.combo_pt[3] = CreateFrame( "Frame", "$parentComboPt"..3, self.frame.target );		
		self.frame.target.combo_pt[4] = CreateFrame( "Frame", "$parentComboPt"..4, self.frame.target );		
		self.frame.target.combo_pt[5] = CreateFrame( "Frame", "$parentComboPt"..5, self.frame.target );		

		for i=1,5 do
			self.frame.target.combo_pt[i].texture = self.frame.target.combo_pt[i]:CreateTexture( "$parentTexture" );
			self.frame.target.combo_pt[i].texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\nUI_ComboPoint.blp" );
		end
		
		self.frame.target.mana.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\rpb.blp" );
		self.frame.target.mana.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );

		self.frame.target.rage.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\rpb.blp" );
		self.frame.target.rage.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.target.energy.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\rpb.blp" );
		self.frame.target.energy.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.target.focus.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\rpb.blp" );
		self.frame.target.focus.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.target.health.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\rhb.blp" );
		self.frame.target.health.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.target.raid_icon.texture:SetTexture( "Interface\\TargetingFrame\\UI-RaidTargetingIcons" );
		
		self.frame.target.powerframe.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\rpf.blp" );
		self.frame.target.powerframe.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.target.healthframe.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\rhf.blp" );
		self.frame.target.healthframe.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.target.unit     = "target";
		self.frame.target.is_shown = true;
		self.frame.target:SetAttribute( "unit", "target" );
		RegisterUnitWatch( self.frame.target );

		-- ToT data elements
		
		self.frame.tot             = CreateFrame( "Frame", "$parent_ToT", self.frame );
		self.frame.tot.health      = CreateFrame( "Frame", "$parentHealth", self.frame.tot );
		self.frame.tot.mana        = CreateFrame( "Frame", "$parentMana", self.frame.tot );
		self.frame.tot.rage        = CreateFrame( "Frame", "$parentRage", self.frame.tot );
		self.frame.tot.energy      = CreateFrame( "Frame", "$parentEnergy", self.frame.tot );
		self.frame.tot.focus       = CreateFrame( "Frame", "$parentFocus", self.frame.tot );
		self.frame.tot.raid_icon   = CreateFrame( "Frame", "$parentRaidIcon", self.frame.tot );
		self.frame.tot.healthframe = CreateFrame( "Frame", "$parentHealthFrame", self.frame.tot );
		self.frame.tot.powerframe  = CreateFrame( "Frame", "$parentPowerFrame", self.frame.tot );

		self.frame.tot.health.texture      = self.frame.tot.health:CreateTexture( "$parentTexture" );
		self.frame.tot.mana.texture        = self.frame.tot.mana:CreateTexture( "$parentTexture" );
		self.frame.tot.rage.texture        = self.frame.tot.rage:CreateTexture( "$parentTexture" );
		self.frame.tot.energy.texture      = self.frame.tot.energy:CreateTexture( "$parentTexture" );
		self.frame.tot.focus.texture       = self.frame.tot.focus:CreateTexture( "$parentTexture" );
		self.frame.tot.raid_icon.texture   = self.frame.tot.raid_icon:CreateTexture( "$parentTexture" );
		self.frame.tot.healthframe.texture = self.frame.tot.healthframe:CreateTexture( "$parentTexture" );
		self.frame.tot.powerframe.texture  = self.frame.tot.powerframe:CreateTexture( "$parentTexture" );
		
		self.frame.tot.health.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\rh2b.blp" );
		self.frame.tot.health.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );

		self.frame.tot.mana.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\rp2b.blp" );
		self.frame.tot.mana.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.tot.rage.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\rp2b.blp" );
		self.frame.tot.rage.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.tot.energy.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\rp2b.blp" );
		self.frame.tot.energy.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.tot.focus.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\rp2b.blp" );
		self.frame.tot.focus.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.tot.raid_icon.texture:SetTexture( "Interface\\TargetingFrame\\UI-RaidTargetingIcons" );
		
		self.frame.tot.healthframe.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\rh2f.blp" );
		self.frame.tot.healthframe.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.tot.powerframe.texture:SetTexture( "Interface\\AddOns\\nUI\\HUD\\Art\\rp2f.blp" );
		self.frame.tot.powerframe.texture:SetTexCoord( 0, 0, 0, 1, 1, 0, 1, 1 );
		
		self.frame.tot.unit     = "targettarget";
		self.frame.tot.is_shown = true;
		self.frame.tot:SetAttribute( "unit", "targettarget" );
		RegisterUnitWatch( self.frame.tot );

		-- misc
		
		self.frame.misc       = {};
		self.frame.misc.range = CreateFrame( "Frame", "$parent_RangeToTarget", self.frame );
		self.frame.cooldowns  = CreateFrame( "Frame", "$parent_Cooldowns", self.parent );

		self.frame.misc.range.text = self.frame.misc.range:CreateFontString( "$parent_Label", "OVERLAY" );
		self.frame.misc.threat = nUI.Class.ThreatBar:new( self.frame, 1, "threat", layout.Bars.Threat2.width );
		self.frame.misc.ktm = nUI.Class.ThreatBar:new( self.frame, 2, "ktm", layout.Bars.KTM.width );
		self.frame.misc.race = nUI.Class.UnitBar_HealthRace:new( self.frame, 1, layout.Bars.HealthRace.width );
	end
	
	self.frame.class = self;
		
	self.frame:SetScale( 1.0 / nUI.scale );
	self.frame:SetFrameStrata( "LOW" );
	self.frame:SetFrameLevel( 0 );
	self.frame:SetWidth( 25 );
	self.frame:SetHeight( 25 );
	self.frame:SetPoint( 
		self.layout.Origin.anchor, nUI_Dashboard, "CENTER", 
		self.layout.Origin.xOfs * nUI.scale, self.layout.Origin.yOfs * nUI.scale
	);

	-- create the player frame
	
	self.frame.player.guid = nil;
	
	self.frame.player:SetFrameLevel( 1 );
	self.frame.player:SetHeight( height );
	self.frame.player:SetWidth( width );
	self.frame.player:SetPoint( "RIGHT", WorldFrame, "CENTER", -self.layout.spacing * nUI.scale, 0 );
	self.frame.player:Hide();
	self.frame.player.power = nil;
	
	-- player frame labels
	
	self:createLabelSet( self.frame.player, "Player", self.layout.Labels.Player, is_init );

	self.frame.player.name = self.frame.player.labels.name.text;
	
	-- player power bars
	
	self.frame.player.mana.max_power = 0;
	self.frame.player.mana.cur_power = 0;
	self.frame.player.mana.max       = self.frame.player.labels.max_power.text;
	self.frame.player.mana.cur       = self.frame.player.labels.cur_power.text;
	self.frame.player.mana.pct       = self.frame.player.labels.pct_power.text;
	
	self.frame.player.mana:SetFrameLevel( 2 );
	self.frame.player.mana:SetWidth( width );
	self.frame.player.mana:SetHeight( height );
	self.frame.player.mana:SetPoint( "BOTTOM", self.frame.player, "BOTTOM", -1, 0 );
	self.frame.player.mana.texture:SetAllPoints( self.frame.player.mana );
	self.frame.player.mana.texture:SetVertexColor( 
		self.layout.Colors.Mana.max.r, 
		self.layout.Colors.Mana.max.g, 
		self.layout.Colors.Mana.max.b,
		1
	);	
	self.frame.player.mana:Hide();

	self.frame.player.rage.max_power = 0;
	self.frame.player.rage.cur_power = 0;
	self.frame.player.rage.max       = self.frame.player.labels.max_power.text;
	self.frame.player.rage.cur       = self.frame.player.labels.cur_power.text;
	self.frame.player.rage.pct       = self.frame.player.labels.pct_power.text;

	self.frame.player.rage:SetFrameLevel( 2 );
	self.frame.player.rage:SetWidth( width );
	self.frame.player.rage:SetHeight( height );
	self.frame.player.rage:SetPoint( "BOTTOM", self.frame.player, "BOTTOM", -1, 0 );
	self.frame.player.rage.texture:SetAllPoints( self.frame.player.rage );
	self.frame.player.rage.texture:SetVertexColor( 
		self.layout.Colors.Rage.max.r, 
		self.layout.Colors.Rage.max.g, 
		self.layout.Colors.Rage.max.b,
		1
	);
	self.frame.player.rage:Hide();

	self.frame.player.energy.max_power = 0;
	self.frame.player.energy.cur_power = 0;
	self.frame.player.energy.max       = self.frame.player.labels.max_power.text;
	self.frame.player.energy.cur       = self.frame.player.labels.cur_power.text;
	self.frame.player.energy.pct       = self.frame.player.labels.pct_power.text;

	self.frame.player.energy:SetFrameLevel( 2 );
	self.frame.player.energy:SetWidth( width );
	self.frame.player.energy:SetHeight( height );
	self.frame.player.energy:SetPoint( "BOTTOM", self.frame.player, "BOTTOM", -1, 0 );
	self.frame.player.energy.texture:SetAllPoints( self.frame.player.energy );
	self.frame.player.energy.texture:SetVertexColor( 
		self.layout.Colors.Energy.max.r, 
		self.layout.Colors.Energy.max.g, 
		self.layout.Colors.Energy.max.b,
		1
	);
	self.frame.player.energy:Hide();

	self.frame.player.focus.max_power = 0;
	self.frame.player.focus.cur_power = 0;
	self.frame.player.focus.max       = self.frame.player.labels.max_power.text;
	self.frame.player.focus.cur       = self.frame.player.labels.cur_power.text;
	self.frame.player.focus.pct       = self.frame.player.labels.pct_power.text;

	self.frame.player.focus:SetFrameLevel( 2 );
	self.frame.player.focus:SetWidth( width );
	self.frame.player.focus:SetHeight( height );
	self.frame.player.focus:SetPoint( "BOTTOM", self.frame.player, "BOTTOM", -1, 0 );
	self.frame.player.focus.texture:SetAllPoints( self.frame.player.focus );
	self.frame.player.focus.texture:SetVertexColor( 
		self.layout.Colors.Focus.max.r, 
		self.layout.Colors.Focus.max.g, 
		self.layout.Colors.Focus.max.b,
		1
	);
	self.frame.player.focus:Hide();

	-- player health bar
	
	self.frame.player.health.max_health = 0;
	self.frame.player.health.cur_health = 0;
	self.frame.player.health.max        = self.frame.player.labels.max_health.text;
	self.frame.player.health.cur        = self.frame.player.labels.cur_health.text;
	self.frame.player.health.pct        = self.frame.player.labels.pct_health.text;

	self.frame.player.health:SetFrameLevel( 2 );
	self.frame.player.health:SetWidth( width );
	self.frame.player.health:SetHeight( height );
	self.frame.player.health:SetPoint( "BOTTOM", self.frame.player, "BOTTOM", -3, 0 );
	self.frame.player.health.texture:SetAllPoints( self.frame.player.health );
	self.frame.player.health.texture:SetVertexColor( 
		self.layout.Colors.Health.max.r, 
		self.layout.Colors.Health.max.g, 
		self.layout.Colors.Health.max.b,
		1
	);

	-- player raid icon
	
	self.frame.player.raid_icon:SetWidth( 30 * self.scale );
	self.frame.player.raid_icon:SetHeight( 30 * self.scale );
	self.frame.player.raid_icon:SetPoint( "CENTER", self.frame.player, "CENTER", -5 * self.scale, 190 * self.scale );
	self.frame.player.raid_icon:Hide();
	
	self.frame.player.raid_icon.texture:SetAllPoints( self.frame.player.raid_icon );

	-- player power bar frame
	
	self.frame.player.powerframe:SetFrameLevel( 3 );
	self.frame.player.powerframe:SetWidth( width );
	self.frame.player.powerframe:SetHeight( height );
	self.frame.player.powerframe:SetPoint( "BOTTOM", self.frame.player, "BOTTOM", -1, 0 );
	self.frame.player.powerframe.texture:SetAllPoints( self.frame.player.powerframe );

	-- player health bar frame
	
	self.frame.player.healthframe:SetFrameLevel( 3 );
	self.frame.player.healthframe:SetWidth( width );
	self.frame.player.healthframe:SetHeight( height );
	self.frame.player.healthframe:SetPoint( "BOTTOM", self.frame.player, "BOTTOM", -3, 0 );
	self.frame.player.healthframe.texture:SetAllPoints( self.frame.player.healthframe );

	-- player debuffs
	
	local hud_layout = self.layout.Buffs.Player;
	local layout = 
	{
		name       = "nUI_HUD_PlayerDebuffs",
		visible    = hud_layout.visible,
		type       = "debuff",
		highlight  = hud_layout.highlight,
		timers     = hud_layout.timers,
		anim       = hud_layout.anim,
		flash      = hud_layout.flash,
		flash_rate = hud_layout.flash_rate,
		
		Shape =
		{
			orient = hud_layout.Shape.orient,
			grow   = hud_layout.Shape.grow,
			rows   = hud_layout.Shape.rows,
			size   = hud_layout.Shape.size * self.scale,
		},
	};
	
	if not self.frame.player.buffs then
		self.frame.player.buffs = nUI.Class.Buffs:new( self.frame.player, self.id, nUI.Class.BuffsLayout:new( layout ), "player" );
	else
		self.frame.player.buffs:layoutFrame( nUI.Class.BuffsLayout:new( layout ) );
	end
	
	self.frame.player.buffs:SetPoint( hud_layout.Origin.anchor, self.frame.player, "CENTER", hud_layout.Origin.xOfs * self.scale, hud_layout.Origin.yOfs * self.scale );
	
	-- create the player pet frame
	
	self.frame.pet.guid = nil;

	self.frame.pet:SetFrameLevel( 1 );
	self.frame.pet:SetHeight( height );
	self.frame.pet:SetWidth( width );
	self.frame.pet:SetPoint( "RIGHT", WorldFrame, "CENTER", -self.layout.spacing * nUI.scale, 0 );
	self.frame.pet:Hide();
	self.frame.pet.power = nil;
	
	-- pet frame labels
	
	self:createLabelSet( self.frame.pet, "Pet", self.layout.Labels.Pet, is_init );

	self.frame.pet.name = self.frame.pet.labels.name.text;
	
	-- pet health bar 

	self.frame.pet.health.max_health = 0;
	self.frame.pet.health.cur_health = 0;
	self.frame.pet.health.max        = self.frame.pet.labels.max_health.text;
	self.frame.pet.health.cur        = self.frame.pet.labels.cur_health.text;
	self.frame.pet.health.pct        = self.frame.pet.labels.pct_health.text;

	self.frame.pet.health:SetFrameLevel( 2 );
	self.frame.pet.health:SetWidth( width );
	self.frame.pet.health:SetHeight( height );
	self.frame.pet.health:SetPoint( "BOTTOM", self.frame.pet, "BOTTOM", 1, 0 );
	self.frame.pet.health.texture:SetAllPoints( self.frame.pet.health );
	self.frame.pet.health.texture:SetVertexColor( 
		self.layout.Colors.Health.max.r, 
		self.layout.Colors.Health.max.g, 
		self.layout.Colors.Health.max.b,
		1
	);

	-- pet power bars
	
	self.frame.pet.mana.max_power = 0;
	self.frame.pet.mana.cur_power = 0;
	self.frame.pet.mana.max       = self.frame.pet.labels.max_power.text;
	self.frame.pet.mana.cur       = self.frame.pet.labels.cur_power.text;
	self.frame.pet.mana.pct       = self.frame.pet.labels.pct_power.text;

	self.frame.pet.mana:SetFrameLevel( 2 );
	self.frame.pet.mana:SetWidth( width );
	self.frame.pet.mana:SetHeight( height );
	self.frame.pet.mana:SetPoint( "BOTTOM", self.frame.pet, "BOTTOM", 3, 0 );
	self.frame.pet.mana.texture:SetAllPoints( self.frame.pet.mana );
	self.frame.pet.mana.texture:SetVertexColor( 
		self.layout.Colors.Mana.max.r, 
		self.layout.Colors.Mana.max.g, 
		self.layout.Colors.Mana.max.b,
		1
	);
	self.frame.pet.mana:Hide();

	self.frame.pet.rage.max_power = 0;
	self.frame.pet.rage.cur_power = 0;
	self.frame.pet.rage.max       = self.frame.pet.labels.max_power.text;
	self.frame.pet.rage.cur       = self.frame.pet.labels.cur_power.text;
	self.frame.pet.rage.pct       = self.frame.pet.labels.pct_power.text;

	self.frame.pet.rage:SetFrameLevel( 2 );
	self.frame.pet.rage:SetWidth( width );
	self.frame.pet.rage:SetHeight( height );
	self.frame.pet.rage:SetPoint( "BOTTOM", self.frame.pet, "BOTTOM", 3, 0 );
	self.frame.pet.rage.texture:SetAllPoints( self.frame.pet.rage );
	self.frame.pet.rage.texture:SetVertexColor( 
		self.layout.Colors.Rage.max.r, 
		self.layout.Colors.Rage.max.g, 
		self.layout.Colors.Rage.max.b,
		1
	);
	self.frame.pet.rage:Hide();

	self.frame.pet.energy.max_power = 0;
	self.frame.pet.energy.cur_power = 0;
	self.frame.pet.energy.max       = self.frame.pet.labels.max_power.text;
	self.frame.pet.energy.cur       = self.frame.pet.labels.cur_power.text;
	self.frame.pet.energy.pct       = self.frame.pet.labels.pct_power.text;

	self.frame.pet.energy:SetFrameLevel( 2 );
	self.frame.pet.energy:SetWidth( width );
	self.frame.pet.energy:SetHeight( height );
	self.frame.pet.energy:SetPoint( "BOTTOM", self.frame.pet, "BOTTOM", 3, 0 );
	self.frame.pet.energy.texture:SetAllPoints( self.frame.pet.energy );
	self.frame.pet.energy.texture:SetVertexColor( 
		self.layout.Colors.Energy.max.r, 
		self.layout.Colors.Energy.max.g, 
		self.layout.Colors.Energy.max.b,
		1
	);
	self.frame.pet.energy:Hide();

	self.frame.pet.focus.max_power = 0;
	self.frame.pet.focus.cur_power = 0;
	self.frame.pet.focus.max       = self.frame.pet.labels.max_power.text;
	self.frame.pet.focus.cur       = self.frame.pet.labels.cur_power.text;
	self.frame.pet.focus.pct       = self.frame.pet.labels.pct_power.text;

	self.frame.pet.focus:SetFrameLevel( 2 );
	self.frame.pet.focus:SetWidth( width );
	self.frame.pet.focus:SetHeight( height );
	self.frame.pet.focus:SetPoint( "BOTTOM", self.frame.pet, "BOTTOM", 3, 0 );
	self.frame.pet.focus.texture:SetAllPoints( self.frame.pet.focus );
	self.frame.pet.focus.texture:SetVertexColor( 
		self.layout.Colors.Focus.max.r, 
		self.layout.Colors.Focus.max.g, 
		self.layout.Colors.Focus.max.b,
		1
	);
	self.frame.pet.focus:Hide();
	
	-- pet raid icon
	
	self.frame.pet.raid_icon:SetWidth( 25 * self.scale );
	self.frame.pet.raid_icon:SetHeight( 25 * self.scale );
	self.frame.pet.raid_icon:SetPoint( "CENTER", self.frame.pet, "CENTER", -10 * self.scale, 10 * self.scale );
	self.frame.pet.raid_icon:Hide();
	
	self.frame.pet.raid_icon.texture:SetAllPoints( self.frame.pet.raid_icon );

	-- pet health bar frame
	
	self.frame.pet.healthframe:SetFrameLevel( 3 );
	self.frame.pet.healthframe:SetWidth( width );
	self.frame.pet.healthframe:SetHeight( height );
	self.frame.pet.healthframe:SetPoint( "BOTTOM", self.frame.pet, "BOTTOM", 1, 0 );
	self.frame.pet.healthframe.texture:SetAllPoints( self.frame.pet.healthframe );

	-- pet power bar frame
	
	self.frame.pet.powerframe:SetFrameLevel( 3 );
	self.frame.pet.powerframe:SetWidth( width );
	self.frame.pet.powerframe:SetHeight( height );
	self.frame.pet.powerframe:SetPoint( "BOTTOM", self.frame.pet, "BOTTOM", 3, 0 );
	self.frame.pet.powerframe.texture:SetAllPoints( self.frame.pet.powerframe );

	-- pet debuffs
	
	local hud_layout = self.layout.Buffs.Pet;
	local layout = 
	{
		name       = "nUI_HUD_PetDebuffs",
		visible    = hud_layout.visible,
		type       = "debuff",
		highlight  = hud_layout.highlight,
		timers     = hud_layout.timers,
		anim       = hud_layout.anim,
		flash      = hud_layout.flash,
		flash_rate = hud_layout.flash_rate,
		
		Shape =
		{
			orient = hud_layout.Shape.orient,
			grow   = hud_layout.Shape.grow,
			rows   = hud_layout.Shape.rows,
			size   = hud_layout.Shape.size * self.scale,
		},
	};

	if not self.frame.pet.buffs then
		self.frame.pet.buffs = nUI.Class.Buffs:new( self.frame.pet, self.id, nUI.Class.BuffsLayout:new( layout ), "pet" );
	else
		self.frame.pet.buffs:layoutFrame( nUI.Class.BuffsLayout:new( layout ) );
	end
	
	self.frame.pet.buffs:SetPoint( hud_layout.Origin.anchor, self.frame.pet, "CENTER", hud_layout.Origin.xOfs * self.scale, hud_layout.Origin.yOfs * self.scale );
	
	-- create the target frame
	
	self.frame.target.guid = nil;

	self.frame.target:SetFrameLevel( 1 );
	self.frame.target:SetHeight( height );
	self.frame.target:SetWidth( width );
	self.frame.target:SetPoint( "LEFT", WorldFrame, "CENTER", self.layout.spacing * nUI.scale, 0 );
	self.frame.target:Hide();
	self.frame.target.power = nil;
	
	-- target frame labels
	
	self:createLabelSet( self.frame.target, "Target", self.layout.Labels.Target, is_init );

	self.frame.target.name = self.frame.target.labels.name.text;
	
	-- target power bars
	
	self.frame.target.mana.max_power = 0;
	self.frame.target.mana.cur_power = 0;
	self.frame.target.mana.max       = self.frame.target.labels.max_power.text;
	self.frame.target.mana.cur       = self.frame.target.labels.cur_power.text;
	self.frame.target.mana.pct       = self.frame.target.labels.pct_power.text;

	self.frame.target.mana:SetFrameLevel( 2 );
	self.frame.target.mana:SetWidth( width );
	self.frame.target.mana:SetHeight( height );
	self.frame.target.mana:SetPoint( "BOTTOM", self.frame.target, "BOTTOM", 1, 0 );
	self.frame.target.mana.texture:SetAllPoints( self.frame.target.mana );
	self.frame.target.mana.texture:SetVertexColor( 
		self.layout.Colors.Mana.max.r, 
		self.layout.Colors.Mana.max.g, 
		self.layout.Colors.Mana.max.b,
		1
	);
	self.frame.target.mana:Hide();

	self.frame.target.rage.max_power = 0;
	self.frame.target.rage.cur_power = 0;
	self.frame.target.rage.max       = self.frame.target.labels.max_power.text;
	self.frame.target.rage.cur       = self.frame.target.labels.cur_power.text;
	self.frame.target.rage.pct       = self.frame.target.labels.pct_power.text;

	self.frame.target.rage:SetFrameLevel( 2 );
	self.frame.target.rage:SetWidth( width );
	self.frame.target.rage:SetHeight( height );
	self.frame.target.rage:SetPoint( "BOTTOM", self.frame.target, "BOTTOM", 1, 0 );
	self.frame.target.rage.texture:SetAllPoints( self.frame.target.rage );
	self.frame.target.rage.texture:SetVertexColor( 
		self.layout.Colors.Rage.max.r, 
		self.layout.Colors.Rage.max.g, 
		self.layout.Colors.Rage.max.b,
		1
	);
	self.frame.target.rage:Hide();

	self.frame.target.energy.max_power = 0;
	self.frame.target.energy.cur_power = 0;
	self.frame.target.energy.max       = self.frame.target.labels.max_power.text;
	self.frame.target.energy.cur       = self.frame.target.labels.cur_power.text;
	self.frame.target.energy.pct       = self.frame.target.labels.pct_power.text;

	self.frame.target.energy:SetFrameLevel( 2 );
	self.frame.target.energy:SetWidth( width );
	self.frame.target.energy:SetHeight( height );
	self.frame.target.energy:SetPoint( "BOTTOM", self.frame.target, "BOTTOM", 1, 0 );
	self.frame.target.energy.texture:SetAllPoints( self.frame.target.energy );
	self.frame.target.energy.texture:SetVertexColor( 
		self.layout.Colors.Energy.max.r, 
		self.layout.Colors.Energy.max.g, 
		self.layout.Colors.Energy.max.b,
		1
	);
	self.frame.target.energy:Hide();

	self.frame.target.focus.max_power = 0;
	self.frame.target.focus.cur_power = 0;
	self.frame.target.focus.max       = self.frame.target.labels.max_power.text;
	self.frame.target.focus.cur       = self.frame.target.labels.cur_power.text;
	self.frame.target.focus.pct       = self.frame.target.labels.pct_power.text;

	self.frame.target.focus:SetFrameLevel( 2 );
	self.frame.target.focus:SetWidth( width );
	self.frame.target.focus:SetHeight( height );
	self.frame.target.focus:SetPoint( "BOTTOM", self.frame.target, "BOTTOM", 1, 0 );
	self.frame.target.focus.texture:SetAllPoints( self.frame.target.focus );
	self.frame.target.focus.texture:SetVertexColor( 
		self.layout.Colors.Focus.max.r, 
		self.layout.Colors.Focus.max.g, 
		self.layout.Colors.Focus.max.b,
		1
	);
	self.frame.target.focus:Hide();

	-- target health bar
	
	self.frame.target.health.max_health = 0;
	self.frame.target.health.cur_health = 0;
	self.frame.target.health.max        = self.frame.target.labels.max_health.text;
	self.frame.target.health.cur        = self.frame.target.labels.cur_health.text;
	self.frame.target.health.pct        = self.frame.target.labels.pct_health.text;

	self.frame.target.health:SetFrameLevel( 2 );
	self.frame.target.health:SetWidth( width );
	self.frame.target.health:SetHeight( height );
	self.frame.target.health:SetPoint( "BOTTOM", self.frame.target, "BOTTOM", 3, 0 );
	self.frame.target.health.texture:SetAllPoints( self.frame.target.health );
	self.frame.target.health.texture:SetVertexColor( 
		self.layout.Colors.Health.max.r, 
		self.layout.Colors.Health.max.g, 
		self.layout.Colors.Health.max.b,
		1
	);

	-- target raid icon
	
	self.frame.target.raid_icon:SetWidth( 30 * self.scale );
	self.frame.target.raid_icon:SetHeight( 30 * self.scale );
	self.frame.target.raid_icon:SetPoint( "CENTER", self.frame.target, "CENTER", 5 * self.scale, 190 * self.scale );
	self.frame.target.raid_icon:Hide();
	
	self.frame.target.raid_icon.texture:SetAllPoints( self.frame.target.raid_icon );
	
	-- combo points
	
	for i=1,5 do
		
		self.frame.target.combo_pt[i]:SetWidth( 22 * self.scale );
		self.frame.target.combo_pt[i]:SetHeight( 22 * self.scale );
		self.frame.target.combo_pt[i]:Hide();
		self.frame.target.combo_pt[i].texture:SetAllPoints( self.frame.target.combo_pt[i] );
		
	end
	
	self.frame.target.combo_pt[1]:SetPoint( "CENTER", self.frame.target, "CENTER", -19 * self.scale, 160 * self.scale );
	self.frame.target.combo_pt[2]:SetPoint( "CENTER", self.frame.target, "CENTER", -10 * self.scale, 140 * self.scale );
	self.frame.target.combo_pt[3]:SetPoint( "CENTER", self.frame.target, "CENTER", -2 * self.scale, 120 * self.scale );
	self.frame.target.combo_pt[4]:SetPoint( "CENTER", self.frame.target, "CENTER", 3 * self.scale, 100 * self.scale );
	self.frame.target.combo_pt[5]:SetPoint( "CENTER", self.frame.target, "CENTER", 7 * self.scale, 80 * self.scale );
	
	-- target power bar frame
	
	self.frame.target.powerframe:SetFrameLevel( 3 );
	self.frame.target.powerframe:SetWidth( width );
	self.frame.target.powerframe:SetHeight( height );
	self.frame.target.powerframe:SetPoint( "BOTTOM", self.frame.target, "BOTTOM", 1, 0 );
	self.frame.target.powerframe.texture:SetAllPoints( self.frame.target.powerframe );

	-- target health bar frame
	
	self.frame.target.healthframe:SetFrameLevel( 3 );
	self.frame.target.healthframe:SetWidth( width );
	self.frame.target.healthframe:SetHeight( height );
	self.frame.target.healthframe:SetPoint( "BOTTOM", self.frame.target, "BOTTOM", 3, 0 );
	self.frame.target.healthframe.texture:SetAllPoints( self.frame.target.healthframe );

	-- target debuffs
	
	local hud_layout = self.layout.Buffs.Target;
	local layout = 
	{
		name       = "nUI_HUD_TargetBuffs",
		visible    = hud_layout.visible,
		type       = "adapt",
		highlight  = hud_layout.highlight,
		timers     = hud_layout.timers,
		anim       = hud_layout.anim,
		flash      = hud_layout.flash,
		flash_rate = hud_layout.flash_rate,
		
		Shape =
		{
			orient = hud_layout.Shape.orient,
			grow   = hud_layout.Shape.grow,
			rows   = hud_layout.Shape.rows,
			size   = hud_layout.Shape.size * self.scale,
		},
	};

	if not self.frame.target.buffs then
		self.frame.target.buffs = nUI.Class.Buffs:new( self.frame.target, self.id, nUI.Class.BuffsLayout:new( layout ), "target" );
	else
		self.frame.target.buffs:layoutFrame( nUI.Class.BuffsLayout:new( layout ) );
	end
	
	self.frame.target.buffs:SetPoint( hud_layout.Origin.anchor, self.frame.target, "CENTER", hud_layout.Origin.xOfs * self.scale, hud_layout.Origin.yOfs * self.scale );
	
	-- create the target of target frame
	
	self.frame.tot.guid = nil;

	self.frame.tot:SetFrameLevel( 1 );
	self.frame.tot:SetHeight( height );
	self.frame.tot:SetWidth( width );
	self.frame.tot:SetPoint( "LEFT", WorldFrame, "CENTER", self.layout.spacing * nUI.scale, 0 );
	self.frame.tot:Hide();
	self.frame.tot.power = nil;
	
	-- tot frame labels
	
	self:createLabelSet( self.frame.tot, "ToT", self.layout.Labels.ToT, is_init );
	
	self.frame.tot.name = self.frame.tot.labels.name.text;
	
	-- tot health bar 

	self.frame.tot.health.max_health = 0;
	self.frame.tot.health.cur_health = 0;
	self.frame.tot.health.max        = self.frame.tot.labels.max_health.text;
	self.frame.tot.health.cur        = self.frame.tot.labels.cur_health.text;
	self.frame.tot.health.pct        = self.frame.tot.labels.pct_health.text;

	self.frame.tot.health:SetFrameLevel( 2 );
	self.frame.tot.health:SetWidth( width );
	self.frame.tot.health:SetHeight( height );
	self.frame.tot.health:SetPoint( "BOTTOM", self.frame.tot, "BOTTOM", -1, 0 );
	self.frame.tot.health.texture:SetAllPoints( self.frame.tot.health );
	self.frame.tot.health.texture:SetVertexColor( 
		self.layout.Colors.Health.max.r, 
		self.layout.Colors.Health.max.g, 
		self.layout.Colors.Health.max.b,
		1
	);

	-- tot power bars
	
	self.frame.tot.mana.max_power = 0;
	self.frame.tot.mana.cur_power = 0;
	self.frame.tot.mana.max       = self.frame.tot.labels.max_power.text;
	self.frame.tot.mana.cur       = self.frame.tot.labels.cur_power.text;
	self.frame.tot.mana.pct       = self.frame.tot.labels.pct_power.text;

	self.frame.tot.mana:SetFrameLevel( 2 );
	self.frame.tot.mana:SetWidth( width );
	self.frame.tot.mana:SetHeight( height );
	self.frame.tot.mana:SetPoint( "BOTTOM", self.frame.tot, "BOTTOM", -3, 0 );
	self.frame.tot.mana.texture:SetAllPoints( self.frame.tot.mana );
	self.frame.tot.mana.texture:SetVertexColor( 
		self.layout.Colors.Mana.max.r, 
		self.layout.Colors.Mana.max.g, 
		self.layout.Colors.Mana.max.b,
		1
	);
	self.frame.tot.mana:Hide();

	self.frame.tot.rage.max_power = 0;
	self.frame.tot.rage.cur_power = 0;
	self.frame.tot.rage.max       = self.frame.tot.labels.max_power.text;
	self.frame.tot.rage.cur       = self.frame.tot.labels.cur_power.text;
	self.frame.tot.rage.pct       = self.frame.tot.labels.pct_power.text;

	self.frame.tot.rage:SetFrameLevel( 2 );
	self.frame.tot.rage:SetWidth( width );
	self.frame.tot.rage:SetHeight( height );
	self.frame.tot.rage:SetPoint( "BOTTOM", self.frame.tot, "BOTTOM", -3, 0 );
	self.frame.tot.rage.texture:SetAllPoints( self.frame.tot.rage );
	self.frame.tot.rage.texture:SetVertexColor( 
		self.layout.Colors.Rage.max.r, 
		self.layout.Colors.Rage.max.g, 
		self.layout.Colors.Rage.max.b,
		1
	);
	self.frame.tot.rage:Hide();

	self.frame.tot.energy.max_power = 0;
	self.frame.tot.energy.cur_power = 0;
	self.frame.tot.energy.max       = self.frame.tot.labels.max_power.text;
	self.frame.tot.energy.cur       = self.frame.tot.labels.cur_power.text;
	self.frame.tot.energy.pct       = self.frame.tot.labels.pct_power.text;

	self.frame.tot.energy:SetFrameLevel( 2 );
	self.frame.tot.energy:SetWidth( width );
	self.frame.tot.energy:SetHeight( height );
	self.frame.tot.energy:SetPoint( "BOTTOM", self.frame.tot, "BOTTOM", -3, 0 );
	self.frame.tot.energy.texture:SetAllPoints( self.frame.tot.energy );
	self.frame.tot.energy.texture:SetVertexColor( 
		self.layout.Colors.Energy.max.r, 
		self.layout.Colors.Energy.max.g, 
		self.layout.Colors.Energy.max.b,
		1
	);
	self.frame.tot.energy:Hide();

	self.frame.tot.focus.max_power = 0;
	self.frame.tot.focus.cur_power = 0;
	self.frame.tot.focus.max       = self.frame.tot.labels.max_power.text;
	self.frame.tot.focus.cur       = self.frame.tot.labels.cur_power.text;
	self.frame.tot.focus.pct       = self.frame.tot.labels.pct_power.text;

	self.frame.tot.focus:SetFrameLevel( 2 );
	self.frame.tot.focus:SetWidth( width );
	self.frame.tot.focus:SetHeight( height );
	self.frame.tot.focus:SetPoint( "BOTTOM", self.frame.tot, "BOTTOM", -3, 0 );
	self.frame.tot.focus.texture:SetAllPoints( self.frame.tot.focus );
	self.frame.tot.focus.texture:SetVertexColor( 
		self.layout.Colors.Focus.max.r, 
		self.layout.Colors.Focus.max.g, 
		self.layout.Colors.Focus.max.b,
		1
	);
	self.frame.tot.focus:Hide();
	
	-- tot raid icon
	
	self.frame.tot.raid_icon:SetWidth( 25 * self.scale );
	self.frame.tot.raid_icon:SetHeight( 25 * self.scale );
	self.frame.tot.raid_icon:SetPoint( "CENTER", self.frame.tot, "CENTER", 10 * self.scale, 10 * self.scale );
	self.frame.tot.raid_icon:Hide();
	
	self.frame.tot.raid_icon.texture:SetAllPoints( self.frame.tot.raid_icon );

	-- tot health bar frame
	
	self.frame.tot.healthframe:SetFrameLevel( 3 );
	self.frame.tot.healthframe:SetWidth( width );
	self.frame.tot.healthframe:SetHeight( height );
	self.frame.tot.healthframe:SetPoint( "BOTTOM", self.frame.tot, "BOTTOM", -1, 0 );
	self.frame.tot.healthframe.texture:SetAllPoints( self.frame.tot.healthframe );

	-- tot power bar frame
	
	self.frame.tot.powerframe:SetFrameLevel( 3 );
	self.frame.tot.powerframe:SetWidth( width );
	self.frame.tot.powerframe:SetHeight( height );
	self.frame.tot.powerframe:SetPoint( "BOTTOM", self.frame.tot, "BOTTOM", -3, 0 );
	self.frame.tot.powerframe.texture:SetAllPoints( self.frame.tot.powerframe );

	-- tot buffs
	
	local hud_layout = self.layout.Buffs.ToT;
	local layout = 
	{
		name       = "nUI_HUD_ToTBuffs",
		visible    = hud_layout.visible,
		type       = "adapt",
		highlight  = hud_layout.highlight,
		timers     = hud_layout.timers,
		anim       = hud_layout.anim,
		flash      = hud_layout.flash,
		flash_rate = hud_layout.flash_rate,
		
		Shape =
		{
			orient = hud_layout.Shape.orient,
			grow   = hud_layout.Shape.grow,
			rows   = hud_layout.Shape.rows,
			size   = hud_layout.Shape.size * self.scale,
		},
	};

	if not self.frame.tot.buffs then
		self.frame.tot.buffs = nUI.Class.Buffs:new( self.frame.tot, self.id, nUI.Class.BuffsLayout:new( layout ), "targettarget" );
	else
		self.frame.tot.buffs:layoutFrame( nUI.Class.BuffsLayout:new( layout ) );
	end
	
	self.frame.tot.buffs:SetPoint( hud_layout.Origin.anchor, self.frame.tot, "CENTER", hud_layout.Origin.xOfs * self.scale, hud_layout.Origin.yOfs * self.scale );
	
	-- create a label frame for the range to target
	
	self.frame.misc.range.text:SetPoint( self.layout.Labels.Range.anchor, self.frame, "CENTER", self.layout.Labels.Range.xOfs * nUI.scale, self.layout.Labels.Range.yOfs * nUI.scale );
	self.frame.misc.range.text:SetFont( "Fonts\\\FRIZQT__.TTF", self.layout.Labels.Range.size * nUI.scale * 1.25, "OUTLINE" );
	self.frame.misc.range.text:SetJustifyH( self.layout.Labels.Range.align );
	self.frame.misc.range.text:SetJustifyV( "MIDDLE" );
	
	self.frame.range = self.frame.misc.range.text;

	-- cooldown HUD
	
	self.frame.cooldowns.first = nil;
	self.frame.cooldowns.count = 0;
	
	self.frame.cooldowns:SetScale( 1.0 / nUI.scale );
	self.frame.cooldowns:SetPoint( self.layout.Bars.Cooldown.anchor, self.frame, "CENTER", self.layout.Bars.Cooldown.xOfs * nUI.scale, self.layout.Bars.Cooldown.yOfs * nUI.scale );
	self.frame.cooldowns:SetHeight( self.layout.Bars.Cooldown.size * nUI.scale );
	self.frame.cooldowns:SetWidth( 200 );
	
	-- player casting bar
	
	if self.layout.Bars.Player.Casting.visible ~= "no" then
		
		self.layout.BarLayouts.PlayerCasting.width  = self.layout.BarLayouts.PlayerCasting.width * nUI.scale;
		self.layout.BarLayouts.PlayerCasting.height = self.layout.BarLayouts.PlayerCasting.height * nUI.scale;
		
		if not self.frame.misc.player_casting then
			self.frame.misc.player_casting = nUI.Class.UnitBar_Casting:new( self.parent, self.id, self.layout.BarLayouts.PlayerCasting, "player" );
		else
			self.frame.misc.player_casting:layoutFrame( self.layout.BarLayouts.PlayerCasting );
		end
			
		self.frame.misc.player_casting:SetPoint( self.layout.Bars.Player.Casting.anchor, self.frame, "CENTER", self.layout.Bars.Player.Casting.xOfs * nUI.scale, self.layout.Bars.Player.Casting.yOfs * nUI.scale );
		self.frame.misc.player_casting:SetScale( 1.0 / nUI.scale );

		self.layout.BarLayouts.PlayerChanneling.width  = self.layout.BarLayouts.PlayerChanneling.width * nUI.scale;
		self.layout.BarLayouts.PlayerChanneling.height = self.layout.BarLayouts.PlayerChanneling.height * nUI.scale;
		
		if not self.frame.misc.player_channel then
			self.frame.misc.player_channel = nUI.Class.UnitBar_Channeling:new( self.parent, self.id, self.layout.BarLayouts.PlayerChanneling, "player" );
		else
			self.frame.misc.player_channel:layoutFrame( self.layout.BarLayouts.PlayerChanneling );
		end
		
		self.frame.misc.player_channel:SetPoint( self.layout.Bars.Player.Casting.anchor, self.frame, "CENTER", self.layout.Bars.Player.Casting.xOfs * nUI.scale, self.layout.Bars.Player.Casting.yOfs * nUI.scale );
		self.frame.misc.player_channel:SetScale( 1.0 / nUI.scale );
		
		if self.layout.Bars.Player.Casting.pet ~= "no" then
			
			self.layout.BarLayouts.PetCasting.width  = self.layout.BarLayouts.PetCasting.width * nUI.scale;
			self.layout.BarLayouts.PetCasting.height = self.layout.BarLayouts.PetCasting.height * nUI.scale;

			if not self.frame.misc.pet_casting then
				self.frame.misc.pet_casting = nUI.Class.UnitBar_Casting:new( self.parent, self.id, self.layout.BarLayouts.PetCasting, "pet" );
			else
				self.frame.misc.pet_casting:layoutFrame( self.layout.BarLayouts.PetCasting );
			end
			
			self.frame.misc.pet_casting:SetPoint( "BOTTOM", self.layout.BarLayouts.PlayerCasting.name, "TOP", 0, 0 );
			self.frame.misc.pet_casting:SetScale( 1.0 / nUI.scale );

			self.layout.BarLayouts.PetChanneling.width  = self.layout.BarLayouts.PetChanneling.width * nUI.scale;
			self.layout.BarLayouts.PetChanneling.height = self.layout.BarLayouts.PetChanneling.height * nUI.scale;

			if not self.frame.misc.pet_channel then
				self.frame.misc.pet_channel = nUI.Class.UnitBar_Channeling:new( self.parent, self.id, self.layout.BarLayouts.PetChanneling, "pet" );
			else
				self.frame.misc.pet_channel:layoutFrame( self.layout.BarLayouts.PetChanneling );
			end
			
			self.frame.misc.pet_channel:SetPoint( "BOTTOM", self.layout.BarLayouts.PlayerCasting.name, "TOP", 0, 0 );
			self.frame.misc.pet_channel:SetScale( 1.0 / nUI.scale );

		end
		
		if self.layout.Bars.Player.Latency.visible ~= "no" then
			
			self.layout.BarLayouts.PlayerLatency.width  = self.layout.BarLayouts.PlayerLatency.width * nUI.scale;
			self.layout.BarLayouts.PlayerLatency.height = self.layout.BarLayouts.PlayerLatency.height * nUI.scale;

			if not self.frame.misc.player_latency then
				self.frame.misc.player_latency = nUI.Class.UnitBar_Latency:new( self.parent, self.id, self.layout.BarLayouts.PlayerLatency, "player" );
			else
				self.frame.misc.player_latency:layoutFrame( self.layout.BarLayouts.PlayerLatency );
			end
			
			self.frame.misc.player_latency:SetPoint( "TOP", self.layout.BarLayouts.PlayerCasting.name, "BOTTOM", 0, 0 );
			self.frame.misc.player_latency:SetScale( 1.0 / nUI.scale );

		end		
	end

	if self.layout.Bars.GCD.shown == "yes" then
		
		if not self.frame.misc.gcd then
			self.frame.misc.gcd = nUI.Class.UnitBar_GCD:new( self.parent, self.id, self.layout.BarLayouts.PlayerCasting.width, self.layout.Bars.GCD.size * nUI.scale, self.layout.Bars.GCD.Color, self.layout.Bars.GCD.latency );
		else
			self.frame.misc.gcd:layoutFrame( self.layout.BarLayouts.PlayerCasting.width, self.layout.Bars.GCD.size * nUI.scale, self.layout.Bars.GCD.Color, self.layout.Bars.GCD.latency );
		end
		
		self.frame.misc.gcd:SetPoint( self.layout.Bars.GCD.anchor, self.frame, "CENTER", self.layout.Bars.GCD.xOfs * nUI.scale, self.layout.Bars.GCD.yOfs * nUI.scale );
		self.frame.misc.gcd:SetScale( 1.0 / nUI.scale );
	end		

	-- target casting bar

	if self.layout.Bars.Target.Casting.visible ~= "no" then

		self.layout.BarLayouts.TargetCasting.width  = self.layout.BarLayouts.TargetCasting.width * nUI.scale;
		self.layout.BarLayouts.TargetCasting.height = self.layout.BarLayouts.TargetCasting.height * nUI.scale;

		if not self.frame.misc.target_casting then
			self.frame.misc.target_casting = nUI.Class.UnitBar_Casting:new( self.parent, self.id, self.layout.BarLayouts.TargetCasting, "target" );
		else
			self.frame.misc.target_casting:layoutFrame( self.layout.BarLayouts.TargetCasting );
		end
		
		self.frame.misc.target_casting:SetPoint( self.layout.Bars.Target.Casting.anchor, self.frame, "CENTER", self.layout.Bars.Target.Casting.xOfs * nUI.scale, self.layout.Bars.Target.Casting.yOfs * nUI.scale );
		self.frame.misc.target_casting:SetScale( 1.0 / nUI.scale );

		self.layout.BarLayouts.TargetChanneling.width  = self.layout.BarLayouts.TargetChanneling.width * nUI.scale;
		self.layout.BarLayouts.TargetChanneling.height = self.layout.BarLayouts.TargetChanneling.height * nUI.scale;

		if not self.frame.misc.target_channel then
			self.frame.misc.target_channel = nUI.Class.UnitBar_Channeling:new( self.parent, self.id, self.layout.BarLayouts.TargetChanneling, "target" );
		else
			self.frame.misc.target_channel:layoutFrame( self.layout.BarLayouts.TargetChanneling );
		end
		
		self.frame.misc.target_channel:SetPoint( self.layout.Bars.Target.Casting.anchor, self.frame, "CENTER", self.layout.Bars.Target.Casting.xOfs * nUI.scale, self.layout.Bars.Target.Casting.yOfs * nUI.scale );
		self.frame.misc.target_channel:SetScale( 1.0 / nUI.scale );
		
	end
	
	-- player casts status bar
	
	local hud_layout = self.layout.Buffs.Casts;
	local layout = 
	{
		name       = "nUI_HUD_PlayerCasts",
		visible    = hud_layout.visible,
		type       = "casts",
		highlight  = hud_layout.highlight,
		timers     = hud_layout.timers,
		anim       = hud_layout.anim,
		flash      = hud_layout.flash,
		flash_rate = hud_layout.flash_rate,
		
		Shape =
		{
			orient = hud_layout.Shape.orient,
			grow   = hud_layout.Shape.grow,
			rows   = hud_layout.Shape.rows,
			size   = hud_layout.Shape.size * nUI.scale,
		},
	};

	if not self.frame.misc.casts then
		self.frame.misc.casts = nUI.Class.Buffs:new( self.frame, self.id, nUI.Class.BuffsLayout:new( layout ), "target" );
	else
		self.frame.misc.casts:layoutFrame( nUI.Class.BuffsLayout:new( layout ) );
	end
	
	self.frame.misc.casts:SetPoint( hud_layout.Origin.anchor, self.frame, "CENTER", hud_layout.Origin.xOfs * nUI.scale, hud_layout.Origin.yOfs * nUI.scale );

	-- health race bar
	
	self.frame.misc.race:SetPoint( "CENTER", self.frame, "CENTER", self.layout.Bars.HealthRace.xOfs * nUI.scale, self.layout.Bars.HealthRace.yOfs * nUI.scale );
	self.frame.misc.race:setEnabled( self.layout.Bars.HealthRace.visible ~= "no" );
	
	-- threat bars

	self.frame.misc.threat:SetPoint( "CENTER", self.frame, "CENTER", self.layout.Bars.Threat2.xOfs * nUI.scale, self.layout.Bars.Threat2.yOfs * nUI.scale );
	self.frame.misc.threat:setEnabled( self.layout.Bars.Threat2.visible ~= "no" );
	
	if not nUI.AddOn.threat then
		self.frame.misc.threat:Hide();
	end

	self.frame.misc.ktm:SetPoint( "CENTER", self.frame, "CENTER", self.layout.Bars.KTM.xOfs * nUI.scale, self.layout.Bars.KTM.yOfs * nUI.scale );
	self.frame.misc.ktm:setEnabled( self.layout.Bars.KTM.visible ~= "no" );

	if not nUI.AddOn.ktm then
		self.frame.misc.ktm:Hide();
	end

	-- and we're done
	
	self.frame:SetAlpha( self.layout.Alpha.rest );
	self.frame:Show();
	
	self:createRangeList();
	self:registerEvents();	

	self:checkGUID( self.frame.player, "player", 0.1640625 );
	self:checkGUID( self.frame.pet, "pet", 0.509765625 );
	self:checkGUID( self.frame.target, "target", 0.1640625 );
	self:checkGUID( self.frame.tot, "targettarget", 0.509765625 );
	
end

-------------------------------------------------------------------------------

function prototype:createLabelSet( frame, focus, config )
		
	frame.labels = {};
	
	local label     = frame.labels;
	local scale     = self.scale;
	local color_set = self.layout.Colors.Health.max;
	
	label.name       = CreateFrame( "Frame", "$parent_"..focus.."_MaxName", frame );
	label.max_health = CreateFrame( "Frame", "$parent_"..focus.."_MaxHealth", frame );
	label.cur_health = CreateFrame( "Frame", "$parent_"..focus.."_CurHealth", frame );
	label.pct_health = CreateFrame( "Frame", "$parent_"..focus.."_PctHealth", frame );
	label.max_power  = CreateFrame( "Frame", "$parent_"..focus.."_MaxPower",  frame );
	label.cur_power  = CreateFrame( "Frame", "$parent_"..focus.."_CurPower",  frame );
	label.pct_power  = CreateFrame( "Frame", "$parent_"..focus.."_PctPower",  frame );

	label.name.text       = label.max_health:CreateFontString( "$parent_Label", "OVERLAY" );
	label.max_health.text = label.max_health:CreateFontString( "$parent_Label", "OVERLAY" );
	label.cur_health.text = label.cur_health:CreateFontString( "$parent_Label", "OVERLAY" );
	label.pct_health.text = label.pct_health:CreateFontString( "$parent_Label", "OVERLAY" );
	label.max_power.text  = label.max_power:CreateFontString(  "$parent_Label",  "OVERLAY" );
	label.cur_power.text  = label.cur_power:CreateFontString(  "$parent_Label",  "OVERLAY" );
	label.pct_power.text  = label.pct_power:CreateFontString(  "$parent_Label",  "OVERLAY" );

	label.max_health.text:SetTextColor( color_set.r, color_set.g, color_set.b, 1 );
	
	label.name.text:SetPoint( config.Name.anchor, frame, "CENTER", config.Name.xOfs * scale, config.Name.yOfs * scale );
	label.name.text:SetFont( "Fonts\\\FRIZQT__.TTF", config.Name.size * scale * 1.25, "OUTLINE" );
	label.name.text:SetJustifyH( config.Name.align );
	label.name.text:SetJustifyV( "MIDDLE" );

	label.max_health.text:SetPoint( config.MaxHealth.anchor, frame, "CENTER", config.MaxHealth.xOfs * scale, config.MaxHealth.yOfs * scale );
	label.max_health.text:SetFont( "Fonts\\\FRIZQT__.TTF", config.MaxHealth.size * scale * 1.25, "OUTLINE" );
	label.max_health.text:SetJustifyH( config.MaxHealth.align );
	label.max_health.text:SetJustifyV( "MIDDLE" );

	label.cur_health.text:SetPoint( config.CurHealth.anchor, frame, "CENTER", config.CurHealth.xOfs * scale, config.CurHealth.yOfs * scale );
	label.cur_health.text:SetFont( "Fonts\\\FRIZQT__.TTF", config.CurHealth.size * scale * 1.25, "OUTLINE" );
	label.cur_health.text:SetJustifyH( config.CurHealth.align );
	label.cur_health.text:SetJustifyV( "MIDDLE" );
	
	label.pct_health.text:SetPoint( config.PctHealth.anchor, frame, "CENTER", config.PctHealth.xOfs * scale, config.PctHealth.yOfs * scale );
	label.pct_health.text:SetFont( "Fonts\\\FRIZQT__.TTF", config.PctHealth.size * scale * 1.25, "OUTLINE" );
	label.pct_health.text:SetJustifyH( config.PctHealth.align );
	label.pct_health.text:SetJustifyV( "MIDDLE" );
	label.pct_health.text:SetVertexColor( 1, 1, 1, 1 );

	label.max_power.text:SetPoint( config.MaxPower.anchor, frame, "CENTER", config.MaxPower.xOfs * scale, config.MaxPower.yOfs * scale );
	label.max_power.text:SetFont( "Fonts\\\FRIZQT__.TTF", config.MaxPower.size * scale * 1.25, "OUTLINE" );
	label.max_power.text:SetJustifyH( config.MaxPower.align );
	label.max_power.text:SetJustifyV( "MIDDLE" );

	label.cur_power.text:SetPoint( config.CurHealth.anchor, frame, "CENTER", config.CurPower.xOfs * scale, config.CurPower.yOfs * scale );
	label.cur_power.text:SetFont( "Fonts\\\FRIZQT__.TTF", config.CurPower.size * scale * 1.25, "OUTLINE" );
	label.cur_power.text:SetJustifyH( config.CurPower.align );
	label.cur_power.text:SetJustifyV( "MIDDLE" );
	
	label.pct_power.text:SetPoint( config.PctPower.anchor, frame, "CENTER", config.PctPower.xOfs * scale, config.PctPower.yOfs * scale );
	label.pct_power.text:SetFont( "Fonts\\\FRIZQT__.TTF", config.PctPower.size * scale * 1.25, "OUTLINE" );
	label.pct_power.text:SetJustifyH( config.PctPower.align );
	label.pct_power.text:SetJustifyV( "MIDDLE" );
	
	if config.Name.shown      == "no" then label.name:Hide(); end;
	if config.MaxHealth.shown == "no" then label.max_health:Hide(); end;
	if config.CurHealth.shown == "no" then label.cur_health:Hide(); end;
	if config.PctHealth.shown == "no" then label.pct_health:Hide(); end;
	if config.MaxPower.shown  == "no" then label.max_power:Hide(); end;
	if config.CurPower.shown  == "no" then label.cur_power:Hide(); end;
	if config.PctPower.shown  == "no" then label.pct_power:Hide(); end;
		
end

-------------------------------------------------------------------------------
-- this method may need a little explanation on intent. It is used to build a
-- list of spells the player has which have a range component to them. That is
-- later used to range check distances to targets. The list is cached for the
-- sake of efficiency and only updated if the player's spellbook changes.

function prototype:createRangeList()

	self.RangeList = {};
	
	local i   = 1;
	local idx = 1;
	
--	nUI:debugMsg( 1, "----- Beginning RangeList Logic -----" );
	
	while true do
		
		-- get a spell from the player's spellbook, quit when we don't find one
		
		local spell, rank = GetSpellName( i, BOOKTYPE_SPELL );
		
		if not spell then break; end
		
		-- determine if the spell has a maximum range and if so, add it to our range list
		
		if SpellHasRange( spell ) then
	
			if not nUI_ExcludedRangedSpells[spell] then
				
				local _, _, _, _, _, _, _, min_range, max_range = GetSpellInfo( spell );
		
				-- insert the new spell into the table at the current index
				
				local new_spell =
				{
					id   = i;
					name = spell,
					rank = rank,
					min  = min_range;
					max  = max_range,
				};
				
				local	exists = self.RangeList[spell];
				
				self.RangeList[spell] = new_spell;
	
				if not exists then
					if-max_range == 0 and not nUI_ZeroRangeSpells[spell] then
--						nUI:debugMsg( 1, "added unknown zero range spell ["..spell.."]" );
					else
--						nUI:debugMsg( 1, "added ranged spell ["..spell.."] range "..min_range.."-"..max_range );				
					end
				end
			end
		else
--			nUI:debugMsg( 1, "skipped unranged spell ["..spell.."]" );			
		end
		
		i = i+1;		
	end
	
--	nUI:debugMsg( 1, "----- Ending RangeList Logic -----" );	
	
end

-------------------------------------------------------------------------------

function prototype:Show()
	
	if not self.frame:IsShown() then
		self.frame:Show();
		self.frame.cooldowns:Show();
	end
end

function prototype:Hide()
	
	if self.frame:IsShown() then
		self.frame:Hide();
		self.frame.cooldowns:Hide();
	end
end

-------------------------------------------------------------------------------

function prototype:updateRaidIcon( frame, unit_id )

	if UnitExists( unit_id ) then
		
		local idx = GetRaidTargetIndex( unit_id );

		if idx == nil then
			
			if frame:IsShown() then
				frame:Hide();
			end
			
		else
			
			local col = ((idx-1) % 4) * 0.25;
			local row = math.floor( (idx-1) / 4 ) * 0.25;
			
			frame.texture:SetTexCoord( col, row, col, row+0.25, col+0.25, row, col+0.25, row+0.25 );
			frame:Show();
			
		end
		
	elseif frame:IsShown() then
		
		frame:Hide();
		
	end	
end

-------------------------------------------------------------------------------
--                              Event Management                             --
-------------------------------------------------------------------------------

function prototype:registerEvents()

	self.frame:SetScript( "OnUpdate", function( who, elapsed ) self:onUpdate( elapsed ); end );
	
	self.frame.player:SetScript( "OnHide", function() self.frame.player.guid = nil; end );	
	self.frame.player:SetScript( "OnShow", function() self:showPlayer(); end );	
	self.frame.player:SetScript( "OnEvent", function() self:playerEvent(); end );
	self.frame.player.health:SetScript( "OnEvent", function() self:playerHealthEvent(); end );
	self.frame.player.mana:SetScript( "OnEvent", function() self:playerManaEvent(); end );
	self.frame.player.rage:SetScript( "OnEvent", function() self:playerRageEvent(); end );
	self.frame.player.focus:SetScript( "OnEvent", function() self:playerFocusEvent(); end );
	self.frame.player.energy:SetScript( "OnEvent", function() self:playerEnergyEvent(); end );
	
	self.frame.pet:SetScript( "OnHide", function() self.frame.pet.guid = nil; end );	
	self.frame.pet:SetScript( "OnShow", function() self:showPet(); end );
	self.frame.pet:SetScript( "OnEvent", function() self:petEvent(); end );
	self.frame.pet.health:SetScript( "OnEvent", function() self:petHealthEvent(); end );
	self.frame.pet.mana:SetScript( "OnEvent", function() self:petManaEvent(); end );
	self.frame.pet.rage:SetScript( "OnEvent", function() self:petRageEvent(); end );
	self.frame.pet.focus:SetScript( "OnEvent", function() self:petFocusEvent(); end );
	self.frame.pet.energy:SetScript( "OnEvent", function() self:petEnergyEvent(); end );
	
	self.frame.target:SetScript( "OnHide", function() self.frame.target.guid = nil; end );	
	self.frame.target:SetScript( "OnShow", function() self:showTarget(); end );
	self.frame.target:SetScript( "OnEvent", function() self:targetEvent(); end );
	self.frame.target.health:SetScript( "OnEvent", function() self:targetHealthEvent(); end );
	self.frame.target.mana:SetScript( "OnEvent", function() self:targetManaEvent(); end );
	self.frame.target.rage:SetScript( "OnEvent", function() self:targetRageEvent(); end );
	self.frame.target.focus:SetScript( "OnEvent", function() self:targetFocusEvent(); end );
	self.frame.target.energy:SetScript( "OnEvent", function() self:targetEnergyEvent(); end );
	
	self.frame.tot:SetScript( "OnHide", function() self.frame.tot.guid = nil; end );	
	self.frame.tot:SetScript( "OnShow", function() self:showToT(); end );
	self.frame.tot:SetScript( "OnEvent", function() self:totEvent(); end );
	self.frame.tot.health:SetScript( "OnEvent", function() self:totHealthEvent(); end );
	self.frame.tot.mana:SetScript( "OnEvent", function() self:totManaEvent(); end );
	self.frame.tot.rage:SetScript( "OnEvent", function() self:totRageEvent(); end );
	self.frame.tot.focus:SetScript( "OnEvent", function() self:totFocusEvent(); end );
	self.frame.tot.energy:SetScript( "OnEvent", function() self:totEnergyEvent(); end );
	
	self.frame.player:RegisterEvent( "UNIT_DISPLAYPOWER" );
	self.frame.player:RegisterEvent( "RAID_TARGET_UPDATE" );
	self.frame.player:RegisterEvent( "LEARNED_SPELL_IN_TAB" );
	self.frame.player.health:RegisterEvent( "UNIT_HEALTH" );
	self.frame.player.health:RegisterEvent( "UNIT_MAXHEALTH" );
	self.frame.player.mana:RegisterEvent( "UNIT_MANA" );
	self.frame.player.mana:RegisterEvent( "UNIT_MAXMANA" );
	self.frame.player.rage:RegisterEvent( "UNIT_RAGE" );
	self.frame.player.rage:RegisterEvent( "UNIT_MAXRAGE" );
	self.frame.player.energy:RegisterEvent( "UNIT_ENERGY" );
	self.frame.player.energy:RegisterEvent( "UNIT_MAXENERGY" );
	self.frame.player.focus:RegisterEvent( "UNIT_FOCUS" );
	self.frame.player.focus:RegisterEvent( "UNIT_MAXFOCUS" );

	self.frame.pet:RegisterEvent( "UNIT_DISPLAYPOWER" );
	self.frame.pet:RegisterEvent( "RAID_TARGET_UPDATE" );
	self.frame.pet:RegisterEvent( "UNIT_PET" );
	self.frame.pet:RegisterEvent( "UNIT_NAME_UPDATE" );
	self.frame.pet.health:RegisterEvent( "UNIT_HEALTH" );
	self.frame.pet.health:RegisterEvent( "UNIT_MAXHEALTH" );
	self.frame.pet.mana:RegisterEvent( "UNIT_MANA" );
	self.frame.pet.mana:RegisterEvent( "UNIT_MAXMANA" );
	self.frame.pet.rage:RegisterEvent( "UNIT_RAGE" );
	self.frame.pet.rage:RegisterEvent( "UNIT_MAXRAGE" );
	self.frame.pet.energy:RegisterEvent( "UNIT_ENERGY" );
	self.frame.pet.energy:RegisterEvent( "UNIT_MAXENERGY" );
	self.frame.pet.focus:RegisterEvent( "UNIT_FOCUS" );
	self.frame.pet.focus:RegisterEvent( "UNIT_MAXFOCUS" );

	self.frame.target:RegisterEvent( "UNIT_DISPLAYPOWER" );
	self.frame.target:RegisterEvent( "RAID_TARGET_UPDATE" );
	self.frame.target:RegisterEvent( "PLAYER_COMBO_POINTS" );
	self.frame.target:RegisterEvent( "UNIT_NAME_UPDATE" );
	self.frame.target:RegisterEvent( "PLAYER_TARGET_CHANGED" );
	self.frame.target.health:RegisterEvent( "UNIT_HEALTH" );
	self.frame.target.health:RegisterEvent( "UNIT_MAXHEALTH" );
	self.frame.target.mana:RegisterEvent( "UNIT_MANA" );
	self.frame.target.mana:RegisterEvent( "UNIT_MAXMANA" );
	self.frame.target.rage:RegisterEvent( "UNIT_RAGE" );
	self.frame.target.rage:RegisterEvent( "UNIT_MAXRAGE" );
	self.frame.target.energy:RegisterEvent( "UNIT_ENERGY" );
	self.frame.target.energy:RegisterEvent( "UNIT_MAXENERGY" );
	self.frame.target.focus:RegisterEvent( "UNIT_FOCUS" );
	self.frame.target.focus:RegisterEvent( "UNIT_MAXFOCUS" );

	self.frame.tot:RegisterEvent( "UNIT_DISPLAYPOWER" );
	self.frame.tot:RegisterEvent( "RAID_TARGET_UPDATE" );
	self.frame.tot:RegisterEvent( "UNIT_NAME_UPDATE" );
	self.frame.tot:RegisterEvent( "PLAYER_TARGET_CHANGED" );
	self.frame.tot.health:RegisterEvent( "UNIT_HEALTH" );
	self.frame.tot.health:RegisterEvent( "UNIT_MAXHEALTH" );
	self.frame.tot.mana:RegisterEvent( "UNIT_MANA" );
	self.frame.tot.mana:RegisterEvent( "UNIT_MAXMANA" );
	self.frame.tot.rage:RegisterEvent( "UNIT_RAGE" );
	self.frame.tot.rage:RegisterEvent( "UNIT_MAXRAGE" );
	self.frame.tot.energy:RegisterEvent( "UNIT_ENERGY" );
	self.frame.tot.energy:RegisterEvent( "UNIT_MAXENERGY" );
	self.frame.tot.focus:RegisterEvent( "UNIT_FOCUS" );
	self.frame.tot.focus:RegisterEvent( "UNIT_MAXFOCUS" );
	
end

-------------------------------------------------------------------------------

function prototype:unregisterEvents()

	self.frame:SetScript( "OnUpdate", nil );
	
	self.frame.player:SetScript( "OnShow", nil );	
	self.frame.player:SetScript( "OnEvent", nil );
	self.frame.player.health:SetScript( "OnEvent", nil );
	self.frame.player.mana:SetScript( "OnEvent", nil );
	self.frame.player.rage:SetScript( "OnEvent", nil );
	self.frame.player.focus:SetScript( "OnEvent", nil );
	self.frame.player.energy:SetScript( "OnEvent", nil );
	
	self.frame.pet:SetScript( "OnShow", nil );
	self.frame.pet:SetScript( "OnEvent", nil );
	self.frame.pet.health:SetScript( "OnEvent", nil );
	self.frame.pet.mana:SetScript( "OnEvent", nil );
	self.frame.pet.rage:SetScript( "OnEvent", nil );
	self.frame.pet.focus:SetScript( "OnEvent", nil );
	self.frame.pet.energy:SetScript( "OnEvent", nil );
	
	self.frame.target:SetScript( "OnShow", nil );
	self.frame.target:SetScript( "OnEvent", nil );
	self.frame.target.health:SetScript( "OnEvent", nil );
	self.frame.target.mana:SetScript( "OnEvent", nil );
	self.frame.target.rage:SetScript( "OnEvent", nil );
	self.frame.target.focus:SetScript( "OnEvent", nil );
	self.frame.target.energy:SetScript( "OnEvent", nil );
	
	self.frame.tot:SetScript( "OnShow", nil );
	self.frame.tot:SetScript( "OnEvent", nil );
	self.frame.tot.health:SetScript( "OnEvent", nil );
	self.frame.tot.mana:SetScript( "OnEvent", nil );
	self.frame.tot.rage:SetScript( "OnEvent", nil );
	self.frame.tot.focus:SetScript( "OnEvent", nil );
	self.frame.tot.energy:SetScript( "OnEvent", nil );
	
	self.frame.player:UnregisterEvent( "UNIT_DISPLAYPOWER" );
	self.frame.player:UnregisterEvent( "RAID_TARGET_UPDATE" );
	self.frame.player:UnregisterEvent( "LEARNED_SPELL_IN_TAB" );
	self.frame.player.health:UnregisterEvent( "UNIT_HEALTH" );
	self.frame.player.health:UnregisterEvent( "UNIT_MAXHEALTH" );
	self.frame.player.mana:UnregisterEvent( "UNIT_MANA" );
	self.frame.player.mana:UnregisterEvent( "UNIT_MAXMANA" );
	self.frame.player.rage:UnregisterEvent( "UNIT_RAGE" );
	self.frame.player.rage:UnregisterEvent( "UNIT_MAXRAGE" );
	self.frame.player.energy:UnregisterEvent( "UNIT_ENERGY" );
	self.frame.player.energy:UnregisterEvent( "UNIT_MAXENERGY" );
	self.frame.player.focus:UnregisterEvent( "UNIT_FOCUS" );
	self.frame.player.focus:UnregisterEvent( "UNIT_MAXFOCUS" );

	self.frame.pet:UnregisterEvent( "UNIT_DISPLAYPOWER" );
	self.frame.pet:UnregisterEvent( "RAID_TARGET_UPDATE" );
	self.frame.pet:UnregisterEvent( "UNIT_NAME_UPDATE" );
	self.frame.pet:UnregisterEvent( "UNIT_PET" );
	self.frame.pet.health:UnregisterEvent( "UNIT_HEALTH" );
	self.frame.pet.health:UnregisterEvent( "UNIT_MAXHEALTH" );
	self.frame.pet.mana:UnregisterEvent( "UNIT_MANA" );
	self.frame.pet.mana:UnregisterEvent( "UNIT_MAXMANA" );
	self.frame.pet.rage:UnregisterEvent( "UNIT_RAGE" );
	self.frame.pet.rage:UnregisterEvent( "UNIT_MAXRAGE" );
	self.frame.pet.energy:UnregisterEvent( "UNIT_ENERGY" );
	self.frame.pet.energy:UnregisterEvent( "UNIT_MAXENERGY" );
	self.frame.pet.focus:UnregisterEvent( "UNIT_FOCUS" );
	self.frame.pet.focus:UnregisterEvent( "UNIT_MAXFOCUS" );

	self.frame.target:UnregisterEvent( "UNIT_DISPLAYPOWER" );
	self.frame.target:UnregisterEvent( "RAID_TARGET_UPDATE" );
	self.frame.target:UnregisterEvent( "PLAYER_COMBO_POINTS" );
	self.frame.target:UnregisterEvent( "PLAYER_TARGET_CHANGED" );
	self.frame.target:UnregisterEvent( "UNIT_NAME_UPDATE" );
	self.frame.target.health:UnregisterEvent( "UNIT_HEALTH" );
	self.frame.target.health:UnregisterEvent( "UNIT_MAXHEALTH" );
	self.frame.target.mana:UnregisterEvent( "UNIT_MANA" );
	self.frame.target.mana:UnregisterEvent( "UNIT_MAXMANA" );
	self.frame.target.rage:UnregisterEvent( "UNIT_RAGE" );
	self.frame.target.rage:UnregisterEvent( "UNIT_MAXRAGE" );
	self.frame.target.energy:UnregisterEvent( "UNIT_ENERGY" );
	self.frame.target.energy:UnregisterEvent( "UNIT_MAXENERGY" );
	self.frame.target.focus:UnregisterEvent( "UNIT_FOCUS" );
	self.frame.target.focus:UnregisterEvent( "UNIT_MAXFOCUS" );

	self.frame.tot:UnregisterEvent( "UNIT_DISPLAYPOWER" );
	self.frame.tot:UnregisterEvent( "RAID_TARGET_UPDATE" );
	self.frame.tot:UnregisterEvent( "UNIT_NAME_UPDATE" );
	self.frame.tot:UnregisterEvent( "PLAYER_TARGET_CHANGED" );
	self.frame.tot.health:UnregisterEvent( "UNIT_HEALTH" );
	self.frame.tot.health:UnregisterEvent( "UNIT_MAXHEALTH" );
	self.frame.tot.mana:UnregisterEvent( "UNIT_MANA" );
	self.frame.tot.mana:UnregisterEvent( "UNIT_MAXMANA" );
	self.frame.tot.rage:UnregisterEvent( "UNIT_RAGE" );
	self.frame.tot.rage:UnregisterEvent( "UNIT_MAXRAGE" );
	self.frame.tot.energy:UnregisterEvent( "UNIT_ENERGY" );
	self.frame.tot.energy:UnregisterEvent( "UNIT_MAXENERGY" );
	self.frame.tot.focus:UnregisterEvent( "UNIT_FOCUS" );
	self.frame.tot.focus:UnregisterEvent( "UNIT_MAXFOCUS" );
	
end

-------------------------------------------------------------------------------

function prototype:onUpdate( elapsed )
	
	self.Timers.update = self.Timers.update + elapsed;
	
	-- update HUD status and visibility
	
	if self.Timers.update >= 0.08 then -- 12.5fps update interval

		self.Timers.update = 0;
		
		-- set an alpha mode for the hud based on the current level of activity

		local alpha;
		
		if InCombatLockdown() then
			alpha = self.layout.Alpha.combat;
		elseif UnitExists( "target" ) and not UnitOnTaxi( "player" ) and not UnitIsDeadOrGhost( "target" ) and (UnitCanAttack( "player", "target" ) or UnitPlayerControlled( "target" ) or self.layout.show_npc ~= "no") then
			alpha = self.layout.Alpha.target;
		elseif UnitIsDeadOrGhost( "player" ) then
			alpha = self.layout.Alpha.rest;
		elseif self.frame.player.buffs and self.frame.player.buffs:iconCount() > 0 then
			alpha = self.layout.Alpha.regen;
		elseif self.frame.player.health.cur_health ~= self.frame.player.health.max_health then
			alpha = self.layout.Alpha.regen;
		elseif self.frame.player.power and self.frame.player.power_type ~= "r" and self.frame.player.power.cur_power ~= self.frame.player.power.max_power then
			alpha = self.layout.Alpha.regen;
		elseif self.frame.player.power and self.frame.player.power_type == "r" and self.frame.player.power.cur_power > 0 then
			alpha = self.layout.Alpha.regen;
		elseif UnitExists( "pet" ) and self.frame.pet.buffs and self.frame.pet.buffs:iconCount() > 0 then
			alpha = self.layout.Alpha.regen;
		elseif UnitExists( "pet" ) and self.frame.pet.health.cur_health ~= self.frame.pet.health.max_health then
			alpha = self.layout.Alpha.regen;
		elseif UnitExists( "pet" ) and self.frame.pet.power and self.frame.pet.power_type ~= "r" and self.frame.pet.power.cur_power ~= self.frame.pet.power.max_power then
			alpha = self.layout.Alpha.regen;
		elseif UnitExists( "pet" ) and self.frame.pet.power and self.frame.pet.power_type == "r" and self.frame.pet.power.cur_power > 0 then
			alpha = self.layout.Alpha.regen;
		else
			alpha = self.layout.Alpha.rest;
		end

		self.frame:SetAlpha( alpha );

		-- if there's nothing in the cooldown frame, hide it
		
		if self.layout.Bars.Cooldown.shown == "no" then
			self.frame.cooldowns:Hide();
		elseif self.frame.cooldowns.count > 0 then
			self.frame.cooldowns:Show();
			self.frame.cooldowns:SetAlpha( max( alpha, self.layout.Alpha.regen ) );
		else
			self.frame.cooldowns:Hide();
		end

		-- whether or not to display the target is a special case based on what type of 
		-- target we are dealing with
		
		if not self.frame.target.is_shown and UnitExists( "target" ) and not UnitIsDeadOrGhost( "target" ) then 
			if UnitCanAttack( "player", "target" ) -- show any unit we can attack
			or UnitPlayerControlled( "target" ) -- show any unit that a player controls
			or self.layout.show_npc ~= "no" then -- otherwise, show NPCs only if player wants them
				self.frame.target.is_shown = true;
				self.frame.target:SetAlpha( 1 );
			end
		elseif self.frame.target.is_shown -- if the frame is being shown
		and (not UnitExists( "target" ) -- and the target doesn't exist
			or UnitIsDeadOrGhost( "target" ) -- or the unit is KIA
			or (self.layout.show_npc == "no" -- or we don't want to show NPCs
				and not UnitCanAttack( "player", "target" ) -- and we cannot attack this unit
				and not UnitPlayerControlled( "target" ) ) ) then -- and this unit is not a player
			self.frame.target.guid = nil;
			self.frame.target.is_shown = false;
			self.frame.target:SetAlpha( 0 );
		end		
		
		-- same rules apply to the target of target except that we never show the 
		-- the TOT if the target isn't shown

		if not self.frame.target.is_shown then
			if self.frame.tot.is_shown then 
				self.frame.tot.guid = nil;
				self.frame.tot.is_shown = false;
				self.frame.tot:SetAlpha( 0 ); 
			end
		elseif not self.frame.tot.is_shown and UnitExists( "targettarget" ) and not UnitIsDeadOrGhost( "targettarget" )then 
			if UnitCanAttack( "player", "targettarget" )
			or UnitPlayerControlled( "targettarget" )
			or self.layout.show_npc ~= "no" then
				self.frame.tot.is_shown = true;
				self.frame.tot:SetAlpha( 1 );
			end
		elseif self.frame.tot.is_shown 
		and (not UnitExists( "targettarget" )
			or UnitIsDeadOrGhost( "targettarget" )
			or (self.layout.show_npc == "no"
				and not UnitCanAttack( "player", "targettarget" )
				and not UnitPlayerControlled( "targettarget" ) ) ) then
			self.frame.tot.guid = nil;
			self.frame.tot.is_shown = false;
			self.frame.tot:SetAlpha( 0 );
		end		
	end
	
	-- update the range to target
	
	self.Timers.range = self.Timers.range + elapsed;
	
	if self.Timers.range >= 0.25 then -- 4fps is plenty fast enough
	
		self.Timers.range = 0;
		
		self:updateRangeToTarget();
		
	end

	-- update the cooldown HUD
	
	self.Timers.cooldown = self.Timers.cooldown + elapsed;
	
	if self.Timers.cooldown > 0.16 then -- 25fps
	
		self.Timers.cooldown = 0;
		
		self:updateCooldowns();
	end
end

-------------------------------------------------------------------------------

function prototype:checkGUID( frame, unit_id, top )
	
	if UnitExists( unit_id ) then
		
		local guid = UnitGUID( unit_id );
		
		if guid ~= frame.guid then
	
			frame.guid = guid;
			
			local localized_class, unit_class = UnitClass( unit_id );
			local color = RAID_CLASS_COLORS[unit_class];
			local name = UnitName( unit_id ) or "";
			
			if not color then
				color =
				{
					r = 1,
					g = 1,
					b = 1,
				}
			end
			
			if name == "Unknown" then name = ""; end
			
			frame.name:SetText( name );
			frame.name:SetTextColor( color.r, color.g, color.b, 1 );
			
			self:updateHealth( frame, unit_id, top );
			self:setPowerBar( frame, unit_id, top );
			self:updateRaidIcon( frame.raid_icon, unit_id );
			self:updateComboPoints( unit_id );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:setPowerBar( frame, unit_id, top )
	
	local pt = UnitPowerType( unit_id );
	local pb = nil;
	
	if pt == 0 then 
		pb = frame.mana;
		frame.power_type = "m";
	elseif pt == 1 then 
		pb = frame.rage;
		frame.power_type = "r";
	elseif pt == 2 then 
		pb = frame.focus;
		frame.power_type = "f";
	elseif pt == 3 then 
		pb = frame.energy;
		frame.power_type = "e";
	end
	
	if pb then
		
		if frame.power then frame.power:Hide(); end
		frame.power = pb;
		
		self:updatePower( frame, unit_id, top, true );
		
		if frame.power.max_power <= 0 then
		
			frame.power:Hide();
			frame.power.max:Hide();
			frame.power.cur:Hide();
			frame.power.pct:Hide();

			if self.layout.show_borders ~= "no" then 
				frame.powerframe.texture:Hide(); 
			end
			
		else
			
			frame.power:Show();
			frame.power.max:Show();
			frame.power.cur:Show();
			frame.power.pct:Show();
			
			if self.layout.show_borders ~= "no" then 
				frame.powerframe.texture:Show(); 
			end
		end
		
	end
end

-------------------------------------------------------------------------------

function prototype:setFrameColor( frame, cur_val, max_val, color_set, top, new_bar )
	
	local dy = cur_val / max_val;
	local r, g, b;
	local low_color, high_color, pct_color;
	local bottom = 0.806640625;
	local height = bottom - top;
	local frame_height = frame:GetParent():GetHeight();
	
	if dy < 0.5 then 
		high_color = color_set.mid;
		low_color  = color_set.min;
		v = dy * 2;
	else
		high_color = color_set.max;
		low_color  = color_set.mid;
		v = (dy - 0.5) * 2;
	end
	
	-- set the frame height and clip
	
	frame:SetHeight( frame_height * (1.0 - bottom + height * dy) );
	frame.texture:SetTexCoord( 0, bottom - height * dy, 0, 1, 1, bottom - height * dy, 1, 1 );
	
	-- set the frame color
	
	r = low_color.r + (high_color.r - low_color.r) * v;
	g = low_color.g + (high_color.g - low_color.g) * v;
	b = low_color.b + (high_color.b - low_color.b) * v;
	
	frame.texture:SetVertexColor( r, g, b, 1 );
	frame.pct:SetTextColor( r, g, b, 1 );
	frame.cur:SetTextColor( r, g, b, 1 );

	if new_bar then
		frame.max:SetTextColor( color_set.max.r, color_set.max.g, color_set.max.b, 1 );
	end
end

-------------------------------------------------------------------------------

function prototype:updateComboPoints( unit_id )
	
	if unit_id == "target" then
		
		local num_points = GetComboPoints();

		for i=1,5 do
			
			if i <= num_points then self.frame.target.combo_pt[i]:Show();
			else  self.frame.target.combo_pt[i]:Hide();
			end
		
		end
	end
end
			
-------------------------------------------------------------------------------
-- this method is called from within the button class as a hook when a cooldown
-- starts, begins to flash or expires so that the HUD can update he user display

function prototype:updateCooldowns()

	local now     = GetTime();
	local size    = self.layout.Bars.Cooldown.size * self.scale;
	local color   = nUI_Config.Colors.CDC1;
	local deletes = {};
	
	-- remove any icons we have that are no longer in the cooldown table
	-- and determine when the countdown ends if it is a new entry
	
	for i in pairs( Cooldowns ) do 
		
		local cooldown = Cooldowns[i];

		-- a new entry in the table
		
		if not cooldown.frame then
			
			cooldown.frame = CreateFrame( "Frame", "$parent_"..cooldown.id, self.frame.cooldowns );
			cooldown.frame.texture = cooldown.frame:CreateTexture( "$parentTexture" );
			cooldown.frame.counter = CreateFrame( "Frame", "$parent_Counter", cooldown.frame );

			cooldown.frame:EnableMouse( true );			
			cooldown.frame:SetHeight( size );
			cooldown.frame:SetWidth( size );
			
			cooldown.frame.texture:SetAllPoints( cooldown.frame );

			if self.layout.Bars.Cooldown.shown ~= "no" then
				
				cooldown.text = cooldown.frame.counter:CreateFontString( "$parentText", "OVERLAY" );
				cooldown.text:SetPoint( "TOPLEFT", cooldown.frame, "BOTTOMLEFT", 0, 1 );
				cooldown.text:SetPoint( "TOPRIGHT", cooldown.frame, "BOTTOMRIGHT", 0, 1 );
				cooldown.text:SetFont( "Fonts\\ARIALN.TTF", size * 0.417, "OUTLINE" );
				cooldown.text:SetTextColor( color.r, color.g, color.b, 1 );
				
			end		
		end

		-- make sure the end time hasn't changed for any reason
		
		cooldown.ends = cooldown.start + cooldown.duration;			

		-- also ensure the icon hasn't changed
		
		local icon = cooldown.icon and cooldown.icon:GetTexture() or nil;
		
		if cooldown.texture ~= icon then
			cooldown.texture = icon;			
			cooldown.frame.texture:SetTexture( icon );
			cooldown.frame:SetScript( "OnEnter", cooldown.onEnter );
			cooldown.frame:SetScript( "OnLeave", cooldown.onLeave );
		end		
	end

	-- sort the table by the time remaining
	
	table.sort( Cooldowns, nUI_SortCooldowns ); 
	
	-- now display all the icons
	
	local width = 0;
	local last  = nil;
	local gap   = self.layout.Bars.Cooldown.gap * self.scale;
	
	self.frame.cooldowns.count = 0;
	
	for i in pairs( Cooldowns ) do
		
		local cooldown = Cooldowns[i];
		local frame    = cooldown.frame;

		if not cooldown.done and cooldown.texture and now >= cooldown.start and now <= cooldown.ends then
				
			frame:ClearAllPoints();
	
			if last then frame:SetPoint( "LEFT", last, "RIGHT", gap, 0 );
			else frame:SetPoint( "LEFT", self.frame.cooldowns, "LEFT", 0, 0 );
			end
			
			if cooldown.flashing then 
				frame.texture:SetVertexColor( 1, 0.5, 0.5, 1 );
			else 
				frame.texture:SetVertexColor( 1, 1, 1, 1 );
			end
			
			frame:Show();
			
			width = width + size + (last and gap or 0);
			last  = frame;
			
			self.frame.cooldowns.count = self.frame.cooldowns.count + 1;
		else 
			
			cooldown.frame:Hide();
			
		end		
	end
	
	self.frame.cooldowns:SetWidth( width );

end

-------------------------------------------------------------------------------

function prototype:updateHealth( frame, unit_id, top )
	
	local text  = "";
	local class = UnitClassification( unit_id );

	frame.health.cur_health, frame.health.max_health = nUI_getHealthData( unit_id );
	
	self:setFrameColor( frame.health, frame.health.cur_health, frame.health.max_health, self.layout.Colors.Health, top );	
	
	frame.health.max:SetText( frame.health.max_health );

	text = frame.health.cur_health;
	
	if class == "elite" or class == "rareelite" then 
		text = text.." (+)"; 
	elseif class == "worldboss" then
		text = text.." (++)";
	end
	
	frame.health.cur:SetText( text );
	frame.health.max:SetText( frame.health.max_health );
	
	local pct = math.floor( frame.health.cur_health / frame.health.max_health * 1000 ) / 10;
	
	if pct == math.floor( pct ) then frame.health.pct:SetText( pct..".0%" );
	else frame.health.pct:SetText( pct.."%" ); end
	
end

-------------------------------------------------------------------------------

function prototype:updatePower( frame, unit_id, top, new_bar )

	local color_set = nil;
	
	frame.power.cur_power = UnitMana( unit_id );
	frame.power.max_power = UnitManaMax( unit_id );

	if frame.power_type == "m"     then color_set = self.layout.Colors.Mana;
	elseif frame.power_type == "r" then color_set = self.layout.Colors.Rage;
	elseif frame.power_type == "e" then color_set = self.layout.Colors.Energy;
	elseif frame.power_type == "f" then color_set = self.layout.Colors.Focus;
	end
	
	if not color_set then
		nUI:debugMsg( 1, "ummm... ooops... ["..frame.power_type.."] is not a valid power type for ["..unit_id.."] -- check HUD code!" );
	else		
		self:setFrameColor( frame.power, frame.power.cur_power, frame.power.max_power, color_set, top, new_bar );
			
		frame.power.max:SetText( frame.power.max_power );
		frame.power.cur:SetText( frame.power.cur_power );
		
		if frame.power.max_power == 0 then
			frame.power.max:SetText( "" );
			frame.power.cur:SetText( "" );
			frame.power.pct:SetText( "" );
		else
			local pct = math.floor( frame.power.cur_power / frame.power.max_power * 1000 ) / 10;
			
			if pct == math.floor( pct ) then frame.power.pct:SetText( pct..".0%" );
			else frame.power.pct:SetText( pct.."%" ); end
		end
	end
	
end

-------------------------------------------------------------------------------

function prototype:updateRangeToTarget()

	local color = self.layout.Colors.Range.oor;
	local text  = "|cFFFFFFFFRANGE:|r ";
	
	-- if we're looking at ourself or have no target, then don't bother
	
	if not UnitExists( "target" ) 
	or UnitIsUnit( "player", "target" ) 
	or UnitIsDeadOrGhost( "target" ) 
	or (self.layout.show_npc == "no" 
		and not UnitPlayerControlled( "target" ) 
		and not UnitCanAttack( "player", "target" ))
	then
		
		text = "";
	
	-- otherwise, if the unit is not visible, we can only say it's more than 100yds away
	
	elseif not UnitIsVisible( "target" ) then
		
		text  = text.."> 100";
		color = self.layout.Colors.Range.oor;

	-- if the unit is visble, then use our range list and interaction info to determine range
	
	else
		
		local min_range = 0;
		local max_range = 100;
		
		-- parse the spell list and see what we learn from it
		
		for i in pairs( self.RangeList ) do
			
			local spell  = self.RangeList[i];
			local id     = spell.id;
			local result = IsSpellInRange( id, BOOKTYPE_SPELL, "target" );
			local usable = SpellStatus[id] and SpellStatus[id].usable;
			
			if result == 1 and usable then
				
				if spell.max < max_range then 
--					nUI:debugMsg( 1, "range update -- ["..spell.name.."] max "..spell.max.." < "..max_range );
					max_range = spell.max; 
				end
				if spell.min > min_range then 
--					nUI:debugMsg( 1, "range update -- ["..spell.name.."] min "..spell.min.." > "..min_range );
					min_range = spell.min; 
				end
				if min_range > max_range then 
--					nUI:debugMsg( 1, "range update -- min "..min_range.." > max "..max_range );
					min_range = 0; 
				end
				
			elseif result == 0 and usable then 
				
				if spell.max > min_range and spell.max < max_range then
--					nUI:debugMsg( 1, "range update -- ["..spell.name.."] max "..spell.max.." > "..min_range.." (oor update)" );
					min_range = spell.max;
				end				
			end
		end	
		
		-- see if we can learn anything narrower from interaction distance
		
		if max_range > 10 and CheckInteractDistance( "target", 3 ) == 1 then
--			nUI:debugMsg( 1, "range update -- target in duel range... max = 10" );
			max_range = 10;
		elseif max_range > 30 and CheckInteractDistance( "target", 4 ) == 1 then			
--			nUI:debugMsg( 1, "range update -- target in follow range... max = 30" );
			max_range = 30;
		end
		
		if min_range < 30 and CheckInteractDistance( "target", 4 ) ~= 1 then
--			nUI:debugMsg( 1, "range update -- target out of follow range... min = 30" );
			min_range = 30;
		elseif min_range < 10 and CheckInteractDistance( "target", 3 ) ~= 1 then
--			nUI:debugMsg( 1, "range update -- target out of duel range... min = 10" );
			min_range = 10;
		end
		
		-- build the range text and color from the result
		
		if max_range <= 5 then text = text.."MELEE";
		elseif min_range == max_range then text = text..min_range;
		else text = text..min_range.."-"..max_range;
		end
		
		if     min_range <= 5  then color = self.layout.Colors.Range.melee;				
		elseif min_range <= 20 then color = self.layout.Colors.Range.short;
		elseif min_range <= 30 then color = self.layout.Colors.Range.med;
		elseif min_range <= 40 then color = self.layout.Colors.Range.long;
		end
		
	end

	-- do the actual update;
	
	self.frame.range:SetTextColor( color.r, color.g, color.b, 1 );
	self.frame.range:SetText( text );
	
	return result;
	
end

-------------------------------------------------------------------------------
-- called from the nUI_Button logic when a new global cooldown starts

function prototype:startGCD( length )
	
	length = length or 1.5;
	
	if self.frame.misc.gcd and length > 0 then self.frame.misc.gcd:startTimer( length ); end
	
end

-------------------------------------------------------------------------------
--                             Player Frame Logic                            --
-------------------------------------------------------------------------------

function prototype:showPlayer()
	
	if self.layout.show_borders == "no" then
		if self.frame.player.healthframe:IsShown() then
			self.frame.player.healthframe.texture:Hide();
			self.frame.player.powerframe.texture:Hide();
		end
	elseif not self.frame.player.healthframe:IsShown() then
		self.frame.player.healthframe:Show();
		if self.frame.player.power and self.frame.player.power.max_power > 0 then
			self.frame.player.powerframe:Show();
		end
	end
	
	self:checkGUID( self.frame.player, "player", 0.1640625 );
	
end

-------------------------------------------------------------------------------

function prototype:playerEvent()
	
	if UnitIsUnit( (arg1 or "player"), "player" ) then
		
		if event == "UNIT_DISPLAYPOWER" then			
			self:setPowerBar( self.frame.player, "player", 0.1640625 );
		elseif event == "LEARNED_SPELL_IN_TAB" then
			self:createRangeList();
		elseif event == "RAID_TARGET_UPDATE" then
			self:updateRaidIcon( self.frame.player.raid_icon, "player" );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:playerHealthEvent()
	
	if UnitIsUnit( arg1, "player" ) then
		
		if event == "UNIT_HEALTH" or event == "UNIT_MAXHEALTH" then
			self:updateHealth( self.frame.player, "player", 0.1640625 );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:playerManaEvent()
	
	if UnitIsUnit( arg1, "player" ) and self.frame.player.power_type == "m" then
		
		if event == "UNIT_MANA" or event == "UNIT_MAXMANA" then
			self:updatePower( self.frame.player, "player", 0.1640625 );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:playerRageEvent()
	
	if UnitIsUnit( arg1, "player" ) and self.frame.player.power_type == "r" then
		
		if event == "UNIT_RAGE" or event == "UNIT_MAXRAGE" then
			self:updatePower( self.frame.player, "player", 0.1640625 );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:playerEnergyEvent()
	
	if UnitIsUnit( arg1, "player" ) and self.frame.player.power_type == "e" then
		
		if event == "UNIT_ENERGY" or event == "UNIT_MAXENERGY" then
			self:updatePower( self.frame.player, "player", 0.1640625 );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:playerFocusEvent()
	
	if UnitIsUnit( arg1, "player" ) and self.frame.player.power_type == "f" then
		
		if event == "UNIT_FOCUS" or event == "UNITMAXFOCUS" then
			self:updatePower( self.frame.player, "player", 0.1640625 );
		end
	end
end


-------------------------------------------------------------------------------
--                                 Pet Frame Logic                           --
-------------------------------------------------------------------------------

function prototype:showPet()
	
	if self.layout.show_borders == "no" then
		if self.frame.pet.healthframe:IsShown() then
			self.frame.pet.healthframe.texture:Hide();
			self.frame.pet.powerframe.texture:Hide();
		end
	elseif not self.frame.pet.healthframe:IsShown() then
		self.frame.pet.healthframe:Show();
		if self.frame.pet.power and self.frame.pet.power.max_power > 0 then
			self.frame.pet.powerframe:Show();
		end
	end
	
	self:checkGUID( self.frame.pet, "pet", 0.509765625 );
	
end

-------------------------------------------------------------------------------

function prototype:petEvent()
	
	if event == "UNIT_PET" and UnitIsUnit( arg1, "player" ) then
		self:checkGUID( self.frame.pet, "pet", 0.509765625 );
		
	elseif event == "UNIT_NAME_UPDATE" then
		self:checkGUID( self.frame.pet, "pet", 0.509765625 );
		
	elseif UnitIsUnit( (arg1 or "pet"), "pet" ) then
		
		if event == "UNIT_DISPLAYPOWER" then
			self:setPowerBar( self.frame.pet, "pet", 0.509765625 );			
		elseif event == "RAID_TARGET_UPDATE" then
			self:updateRaidIcon( self.frame.pet.raid_icon, "pet" );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:petHealthEvent()
	
	if UnitIsUnit( arg1, "pet" ) then
		
		if event == "UNIT_HEALTH" or event == "UNIT_MAXHEALTH" then
			self:updateHealth( self.frame.pet, "pet", 0.509765625 );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:petManaEvent()
	
	if UnitIsUnit( arg1, "pet" ) and self.frame.pet.power_type == "m" then
		
		if event == "UNIT_MANA" or event == "UNIT_MAXMANA" then
			self:updatePower( self.frame.pet, "pet", 0.509765625 );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:petRageEvent()
	
	if UnitIsUnit( arg1, "pet" ) and self.frame.pet.power_type == "r" then
		
		if event == "UNIT_RAGE" or event == "UNIT_MAXRAGE" then
			self:updatePower( self.frame.pet, "pet", 0.509765625 );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:petEnergyEvent()
	
	if UnitIsUnit( arg1, "pet" ) and self.frame.pet.power_type == "e" then
		
		if event == "UNIT_ENERGY" or event == "UNIT_MAXENERGY" then
			self:updatePower( self.frame.pet, "pet", 0.509765625 );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:petFocusEvent()
	
	if UnitIsUnit( arg1, "pet" ) and self.frame.pet.power_type == "f" then
		
		if event == "UNIT_FOCUS" or event == "UNITMAXFOCUS" then
			self:updatePower( self.frame.pet, "pet", 0.509765625 );
		end
	end
end

-------------------------------------------------------------------------------
--                              Target Frame Logic                           --
-------------------------------------------------------------------------------

function prototype:showTarget()
	
	if self.layout.show_borders == "no" then
		if self.frame.target.healthframe:IsShown() then
			self.frame.target.healthframe.texture:Hide();
			self.frame.target.powerframe.texture:Hide();
		end
	elseif not self.frame.target.healthframe:IsShown() then
		self.frame.target.healthframe:Show();
		if self.frame.target.power and self.frame.target.power.max_power > 0 then
			self.frame.target.powerframe:Show();
		end
	end

	self:checkGUID( self.frame.target, "target", 0.1640625 );
	
end

-------------------------------------------------------------------------------

function prototype:targetEvent()

	if event == "UNIT_NAME_UPDATE" then
		self:checkGUID( self.frame.target, "target", 0.1640625 );
		
	elseif event == "PLAYER_TARGET_CHANGED" then
		self:checkGUID( self.frame.target, "target", 0.1640625 );
		
	elseif UnitIsUnit( (arg1 or "target"), "target" ) then
		
		if event == "UNIT_DISPLAYPOWER" then
			self:setPowerBar( self.frame.target, "target", 0.1640625 );
		elseif event == "RAID_TARGET_UPDATE" then
			self:updateRaidIcon( self.frame.target.raid_icon, "target" );
		elseif event == "PLAYER_COMBO_POINTS" then
			self:updateComboPoints( "target" );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:targetHealthEvent()
	
	if UnitIsUnit( arg1, "target" ) then
		
		if event == "UNIT_HEALTH" or event == "UNIT_MAXHEALTH" then
			self:updateHealth( self.frame.target, "target", 0.1640625 );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:targetManaEvent()
	
	if UnitIsUnit( arg1, "target" ) and self.frame.target.power_type == "m" then
		
		if event == "UNIT_MANA" or event == "UNIT_MAXMANA" then
			self:updatePower( self.frame.target, "target", 0.1640625 );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:targetRageEvent()
	
	if UnitIsUnit( arg1, "target" ) and self.frame.target.power_type == "r" then
		
		if event == "UNIT_RAGE" or event == "UNIT_MAXRAGE" then
			self:updatePower( self.frame.target, "target", 0.1640625 );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:targetEnergyEvent()
	
	if UnitIsUnit( arg1, "target" ) and self.frame.target.power_type == "e" then
		
		if event == "UNIT_ENERGY" or event == "UNIT_MAXENERGY" then
			self:updatePower( self.frame.target, "target", 0.1640625 );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:targetFocusEvent()
	
	if UnitIsUnit( arg1, "target" ) and self.frame.target.power_type == "f" then
		
		if event == "UNIT_FOCUS" or event == "UNITMAXFOCUS" then
			self:updatePower( self.frame.target, "target", 0.1640625 );
		end
	end
end

-------------------------------------------------------------------------------
--                         Target of Target Frame Logic                      --
-------------------------------------------------------------------------------

function prototype:showToT()
	
	if self.layout.show_borders == "no" then
		if self.frame.tot.healthframe:IsShown() then
			self.frame.tot.healthframe.texture:Hide();
			self.frame.tot.powerframe.texture:Hide();
		end
	elseif not self.frame.tot.healthframe:IsShown() then
		self.frame.tot.healthframe:Show();
		if self.frame.tot.power and self.frame.tot.power.max_power > 0 then
			self.frame.tot.powerframe:Show();
		end
	end
	
	self:checkGUID( self.frame.tot, "targettarget", 0.509765625 );
	
end

-------------------------------------------------------------------------------

function prototype:totEvent()
	
	if event == "UNIT_NAME_UPDATE" then
		self:checkGUID( self.frame.tot, "targettarget", 0.509765625 );
		
	elseif event == "PLAYER_TARGET_CHANGED" then
		self:checkGUID( self.frame.tot, "targettarget", 0.509765625 );
		
	elseif UnitIsUnit( (arg1 or "targettarget"), "targettarget" ) then
		
		if event == "UNIT_DISPLAYPOWER" then
			self:setPowerBar( self.frame.tot, "targettarget", 0.509765625 );
		elseif event == "RAID_TARGET_UPDATE" then
			self:updateRaidIcon( self.frame.tot.raid_icon, "targettarget" );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:totHealthEvent()
	
	if UnitIsUnit( arg1, "targettarget" ) then
		
		if event == "UNIT_HEALTH" or event == "UNIT_MAXHEALTH" then
			self:updateHealth( self.frame.tot, "targettarget", 0.509765625 );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:totManaEvent()
	
	if UnitIsUnit( arg1, "targettarget" ) and self.frame.tot.power_type == "m" then
		
		if event == "UNIT_MANA" or event == "UNIT_MAXMANA" then
			self:updatePower( self.frame.tot, "targettarget", 0.509765625 );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:totRageEvent()
	
	if UnitIsUnit( arg1, "targettarget" ) and self.frame.tot.power_type == "r" then
		
		if event == "UNIT_RAGE" or event == "UNIT_MAXRAGE" then
			self:updatePower( self.frame.tot, "targettarget", 0.509765625 );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:totEnergyEvent()
	
	if UnitIsUnit( arg1, "targettarget" ) and self.frame.tot.power_type == "e" then
		
		if event == "UNIT_ENERGY" or event == "UNIT_MAXENERGY" then
			self:updatePower( self.frame.tot, "targettarget", 0.509765625 );
		end
	end
end

-------------------------------------------------------------------------------

function prototype:totFocusEvent()
	
	if UnitIsUnit( arg1, "targettarget" ) and self.frame.tot.power_type == "f" then
		
		if event == "UNIT_FOCUS" or event == "UNITMAXFOCUS" then
			self:updatePower( self.frame.tot, "targettarget", 0.509765625 );
		end
	end
end

